from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

import pandas as pd
from pydantic import UUID4, BaseModel, Field, model_validator
from pydantic.networks import IPvAnyAddress
from typing_extensions import Annotated

from pypsse.common import CASESTUDY_FOLDER, EXPORTS_FOLDER, LOGS_FOLDER
from pypsse.enumerations import (
    ApiCommands,
    AreaProperties,
    BranchProperties,
    BusProperties,
    ChannelTypes,
    DCLineProperties,
    ExportModes,
    FixedShuntProperties,
    HelicsCoreTypes,
    InductionGeneratorProperties,
    LoadProperties,
    LoggingLevels,
    MachinesProperties,
    ModelProperties,
    ModelTypes,
    ProjectFolders,
    SimulationModes,
    StationProperties,
    SubscriptionFileRequiredColumns,
    SwitchedShuntProperties,
    TransformerProperties,
    UseModes,
    WritableModelTypes,
    ZoneProperties,
)



class SimSettings(BaseModel):
    "Simulation setting  model defination"
    simulation_time: timedelta = timedelta(seconds=3.0)
    simulation_step_resolution: timedelta = timedelta(seconds=0.025)
    psse_solver_timestep: timedelta = timedelta(seconds=0.00333333333)
    start_time: datetime = "01/01/2020 00:00:00"
    use_profile_manager: bool
    project_path: Path = "C:/Users/alatif/Desktop/NAERM/models/pyPSSEModel"
    case_study: Optional[Path] = None
    raw_file: Optional[Path] = None
    snp_file: Optional[Path] = None
    dyr_file: Optional[Path] = None
    rwm_file: Optional[Path] = None
    gic_file: Optional[Path] = None
    subscriptions_file: Optional[Path] = None
    user_models: List[str] = []
    setup_files: List[str] = []
    simulation_mode: SimulationModes

    @model_validator(mode="after")
    def sim_res_smaller_than_sim_time(self):
        assert (
            self.simulation_step_resolution <= self.simulation_time
        ), "simulation_step_resolution should be smaller than simulation_time"
        return self

    @model_validator(mode="after")
    def psse_res_smaller_than_sim_time(self):
        assert (
            self.psse_solver_timestep <= self.simulation_time
        ), "psse_solver_timestep should be smaller than simulation_time"
        return self

    @model_validator(mode="after")
    def validate_case_study(self):
        file_types = [
            "case_study",
            "raw_file",
            "snp_file",
            "dyr_file",
            "rwm_file",
            "gic_file",
        ]
        base_project_path = self.project_path
        for file in file_types:
            file_path = getattr(self, file)
            if (
                file_path
                and str(file_path) != "."
                and not Path(file_path).exists()
            ):
                file_path = base_project_path / CASESTUDY_FOLDER / file_path
                setattr(self, file, file_path)
                assert file_path.exists(), f"{file_path} does not exist"
        return self

    @model_validator(mode="after")
    def validate_subscription_file(self):
        base_project_path = self.project_path
        if self.subscriptions_file and str(self.subscriptions_file) != ".":
            self.subscriptions_file = (
                base_project_path / self.subscriptions_file
            )
            assert (
                self.subscriptions_file.exists()
            ), f"{self.subscriptions_file} does not exist"
            data = pd.read_csv(self.subscriptions_file)
            csv_cols = set(data.columns)
            sub_cols = {e.value for e in SubscriptionFileRequiredColumns}
            assert sub_cols.issubset(
                csv_cols
            ), f"{sub_cols} are required columns for a valid subscription file"
        return self

    @model_validator(mode="after")
    def validate_user_models(self):
        base_project_path = self.project_path
        if self.user_models:
            paths = []
            for file in self.user_models:
                model_file = base_project_path / CASESTUDY_FOLDER / file
                assert model_file.exists(), f"{model_file} does not esist"
                assert (
                    model_file.suffix == ".dll"
                ), "Invalid file extension. Use dll files"
                paths.append(model_file)
            self.user_models = paths
        return self

    @model_validator(mode="after")
    def validate_simulation_mode(self):
        if self.simulation_mode in [
            SimulationModes.DYNAMIC,
            SimulationModes.SNAP,
        ]:
            assert (
                not self.use_profile_manager
            ), "Profile manager can not be used for dynamic simulations. Set 'Use profile manager' to False"
        return self


class ExportSettings(BaseModel):
    "Export setting model defination"

    outx_file: Path = "test.outx"
    out_file: Path = "20LS11p.out"
    excel_file: Path = "20LS11p.xls"
    log_file: Path = "20LS11p.log"
    coordinate_file: Path = ""
    networkx_graph_file: Path = "20LS11p.gpickle"


class PublicationDefination(BaseModel):
    """Publication setting model defination"""

    bus_subsystems: List[int] = [
        0,
    ]
    asset_type: ModelTypes = "buses"
    asset_properties: List[ModelProperties] = ["FREQ", "PU"]


class HelicsSettings(BaseModel):
    "HELICS co-simualtion setting model defination"
    cosimulation_mode: bool = False
    federate_name: str = "psse"
    time_delta: timedelta = timedelta(seconds=0.00333333333)
    core_type: HelicsCoreTypes = "zmq"
    uninterruptible: bool = True
    helics_logging_level: int = Field(5, ge=1, le=10)
    create_subscriptions: bool = True
    iterative_mode: bool = False
    error_tolerance: float = Field(1e-5, g=0)
    max_coiterations: int = Field(15, ge=1)
    broker_ip: IPvAnyAddress = "127.0.0.1"
    broker_port: int = 23404
    disable_generation_on_coupled_buses: bool = True
    publications: List[PublicationDefination]


class LogSettings(BaseModel):
    "Logger setting model defination"
    disable_psse_logging: bool = True
    logging_level: LoggingLevels = LoggingLevels.DEBUG
    log_to_external_file: bool = True
    display_on_screen: bool = True
    clear_old_log_file: bool = True
    pre_configured_logging: bool = False


class PlotSettings(BaseModel):
    "Plotting setting model defination"
    enable_dynamic_plots: bool = False
    columns: int = Field(15, ge=1)


class GICExportSettings(BaseModel):
    "GIC export setting model defination"
    include_substation_connections: bool = False
    include_transfomer_connections: bool = False
    include_branch_connections: bool = True


class BusSubsystems(BaseModel):
    "Bus subsystem model defination"
    from_file: bool = False
    bus_file: Optional[str] = None
    bus_subsystem_list: List[List[int]] = [
        [
            74012,
            17735,
            20115,
            38205,
            70008,
            80511,
        ],
    ]
    publish_subsystems: List[int] = [
        0,
    ]


class LoadBreakdown(BaseModel):
    "Load ZIP model defination"
    constant_current_percentage: float = 0.0
    constant_admittance_percentage: float = 0.0


class LoadSettings(BaseModel):
    "Load model defination"
    convert: bool = True
    active_load: LoadBreakdown = LoadBreakdown()
    reactive_load: LoadBreakdown = LoadBreakdown()


class GeneratorSettings(BaseModel):
    "Generator model defination"
    missing_machine_model: int = 1


class BusFault(BaseModel):
    "Bus fault model defination"
    time: float = 0.2
    bus_id: int = 38205
    duration: float = 0.3
    bus_trip: bool = False
    trip_delay: float = 0.05
    fault_impedance: List[float] = [
        1.0,
        1.0,
    ]


class BusTrip(BaseModel):
    "Bus trip model defination"
    time: float = 0.2
    bus_id: int = 38205


class LineFault(BaseModel):
    "Line fault model defination"
    time: float = 0.2
    bus_ids: List[int]
    duration: float = 0.3
    bus_trip: bool = False
    trip_delay: float = 0.05
    fault_impedance: List[float] = [
        1.0,
        1.0,
    ]


class LineTrip(BaseModel):
    "Line trip model defination"
    time: float = 0.2
    bus_ids: List[int]


class MachineTrip(BaseModel):
    "Machine trip model defination"
    time: float = 0.2
    bus_id: int = 38205
    machine_id: str = ""


class SimulationSettings(BaseModel):
    "PyPSSE project settings"

    simulation: SimSettings
    export: ExportSettings = ExportSettings()
    helics: Optional[HelicsSettings] = None
    log: LogSettings = LogSettings()
    plots: Optional[PlotSettings] = None
    gic_export_settings: Optional[GICExportSettings] = None
    bus_subsystems: BusSubsystems = BusSubsystems()
    loads: LoadSettings = LoadSettings()
    generators: GeneratorSettings = GeneratorSettings()
    contingencies: Optional[
        List[Union[BusFault, LineFault, LineTrip, BusTrip, MachineTrip]]
    ] = None

    @model_validator(mode="after")
    def validate_export_paths(self):
        base_project_path = self.simulation.project_path
        if self.export.outx_file:
            self.export.outx_file = (
                base_project_path / EXPORTS_FOLDER / self.export.outx_file
            )
        if self.export.out_file:
            self.export.out_file = (
                base_project_path / EXPORTS_FOLDER / self.export.out_file
            )
        if self.export.excel_file:
            self.export.excel_file = (
                base_project_path / EXPORTS_FOLDER / self.export.excel_file
            )
        if self.export.log_file:
            self.export.log_file = (
                base_project_path / LOGS_FOLDER / self.export.log_file
            )
        if self.export.networkx_graph_file:
            self.export.networkx_graph_file = (
                base_project_path
                / EXPORTS_FOLDER
                / self.export.networkx_graph_file
            )
        if self.export.coordinate_file:
            self.export.coordinate_file = (
                base_project_path / EXPORTS_FOLDER / self.export.coordinate_file
            )
        return self


class BusChannel(BaseModel):
    "Bus channel model defination"

    asset_type: Literal[ChannelTypes.BUSES.value]
    use: UseModes = UseModes.LIST
    regex: str = ""
    asset_list: List[int] = []
    asset_properties: List[str] = ["voltage_and_angle", "frequency"]


class LoadChannel(BaseModel):
    "Load channel model defination"

    asset_type: Literal[ChannelTypes.LOADS.value]
    use: UseModes = UseModes.LIST
    regex: str = ""
    asset_list: List[List[str]] = [[]]
    asset_properties: List[str] = []


class MachineChannel(BaseModel):
    "Machine channel model defination"

    asset_type: Literal[ChannelTypes.MACHINES.value]
    use: UseModes = UseModes.LIST
    regex: str = ""
    asset_list: List[List[str]] = [[]]
    asset_properties: List[str] = ["PELEC", "QELEC", "SPEED"]


channel_types = Annotated[
    Union[BusChannel, LoadChannel, MachineChannel],
    Field(discriminator="asset_type"),
]


class ExportAssetTypes(BaseModel):
    "Valid export models and associated options"

    buses: Optional[List[BusProperties]] = None
    areas: Optional[List[AreaProperties]] = None
    zones: Optional[List[ZoneProperties]] = None
    stations: Optional[List[StationProperties]] = None
    dctransmissionlines: Optional[List[DCLineProperties]] = None
    loads: Optional[List[LoadProperties]] = None
    fixed_shunts: Optional[List[FixedShuntProperties]] = None
    switched_shunts: Optional[List[SwitchedShuntProperties]] = None
    transformers: Optional[List[TransformerProperties]] = None
    branches: Optional[List[BranchProperties]] = None
    induction_generators: Optional[List[InductionGeneratorProperties]] = None
    machines: Optional[List[MachinesProperties]] = None


class ExportFileOptions(ExportAssetTypes):
    "Export settings for a PyPSSE project"

    export_results_using_channels: bool = False
    defined_subsystems_only: bool = True
    file_format: ExportModes = "h5"
    channels: Optional[List[str]] = None
    channel_setup: Optional[List[channel_types]] = None


class ProjectDefination(BaseModel):
    "PyPSSE project defination"
    overwrite: bool = False
    autofill: bool = True
    project_name: str
    project_folders: List[ProjectFolders] = list(ProjectFolders)
    simulation_settings: SimulationSettings
    export_settings: ExportFileOptions


class MdaoOutput(BaseModel):
    buses: List[int]
    quantities: Dict[ModelTypes, List[str]]


class MdaoInput(BaseModel):
    asset_type: WritableModelTypes
    asset_bus: int
    asset_id: str
    attributes: Dict


class MdaoProblem(BaseModel):
    outputs: MdaoOutput
    inputs: List[MdaoInput]


class ApiPsseReply(BaseModel):
    status: str
    message: Union[Dict, str]
    uuid: Union[UUID4, None] = None


class ApiPsseException(BaseModel):
    message: str
    uuid: Union[UUID4, None] = None


class ApiPsseReplyInstances(BaseModel):
    status: str
    message: str
    simulators: List[UUID4] = []


class ApiAssetQuery(BaseModel):
    asset_type: ModelTypes
    asset_property: Optional[
        Union[
            BusProperties,
            AreaProperties,
            ZoneProperties,
            StationProperties,
            DCLineProperties,
            LoadProperties,
            FixedShuntProperties,
            SwitchedShuntProperties,
            TransformerProperties,
            BranchProperties,
            InductionGeneratorProperties,
            MachinesProperties,
        ]
    ] = None
    asset_id: Optional[str] = None

    @model_validator(mode="after")
    def define_atleast_one(self):
        assert not (
            self.asset_id is None and self.asset_property is None
        ), "Atleast one 'asset_id' or 'asset_property' should be defined"
        return self


class ApiPssePutRequest(BaseModel):
    uuid: UUID4
    command: ApiCommands
    parameters: Optional[ApiAssetQuery] = None


class ApiPssePostRequest(BaseModel):
    project_name: str = "static_example"


class ApiWebSocketRequest(BaseModel):
    command: ApiCommands
    parameters: Optional[Dict] = None


class Contingencies(BaseModel):
    contingencies: List[Union[BusFault, BusTrip, LineFault, LineTrip, MachineTrip]]



"""Main CLI command for PyDSS."""

import click
from loguru import logger

from pypsse.cli.create_profiles import create_profiles
from pypsse.cli.create_project import create_project
from pypsse.cli.run import run

server_dependencies_installed = True

try:
    from pypsse.cli.serve import serve
except ImportError:
    server_dependencies_installed = False
    logger.warning(
        "Server dependencies not installed. Use 'pip install pssepy[server]' to install additonal dependencies"
    )


@click.group()
def cli():
    """PyPSSE commands"""


cli.add_command(create_project)
cli.add_command(run)
cli.add_command(create_profiles)
if server_dependencies_installed:
    cli.add_command(serve)



# --------------------------------------------------------------------------------------------------
# DC-AC Tool: main function
# Bin Wang
# 4/28/2022
#
# Reference:
# Wang, Bin, and Jin Tan. 2022. DC-AC Tool: Fully Automating the Acquisition of AC Power Flow
# Solution. Golden, CO: National Renewable Energy Laboratory. NREL/TP-6A40-80100.
# https://www.nrel.gov/docs/fy22osti/80100.pdf
# --------------------------------------------------------------------------------------------------


import random

import numpy as np
from loguru import logger

from pypsse.utils.dc2ac.helper_functions import PowerFlowData


class DC2ACconverter:
    min_kv_filter = 200.0
    filter_buses = 1.0

    def __init__(self, psse, solver, settings, raw_data) -> None:
        self.settings = settings
        self.raw_data = raw_data
        self.solver = solver
        self.psse = psse
        self._f = psse._f
        self._i = psse._i
        self._s = psse._s
        self.converged = False

    def run(self):
        self.pfd = PowerFlowData(self.psse)  # original power flow data
        self.pfdt = PowerFlowData(self.psse)
        self.all_subs = self.get_bus_list()

        self.pfd.getdata()  # get power flow data
        solved_flag = self.if_solved(1, 5)  ## Step 1
        self.sol_out_orig(solved_flag)

        # if directly solved, save the solved case and move to the next
        logger.info(f"solved_flag:{solved_flag}")
        if solved_flag == 0:  ## IF-1, yes
            self.converged = True
            return

        ## IF-1, no
        # try to solve the unsolved AC power flow
        # add generators
        ad_pq_busid, ad_pv_busid = self.add_generators()  ## Step 2

        # solve 1st power flow
        solved_flag_1st = self.if_solved(0, 5)  ## Step 3, IF-2

        # if 1st power flow not converging, check solvability
        if solved_flag_1st:  ## IF-2, no
            logger.info("First power flow (with added generators) cannot converge! Investigating solvability...")
            self.solver.reload()
            insolvable_flag, dec_perc = self.svblt_check()  ## Step 4 - Step 6

            if insolvable_flag == 1:  # still cannot solve  ## Possibility c,
                self.converged = False
                return

            # try to approach the target loading
            step2_tempcase = self.save_raw(4, 2)
            inc_cur, inc_target, inc_step = self.approach_target_loading(dec_perc, step2_tempcase)  ## Step 5

            # Recover the original loading and dispatch
            if inc_cur >= inc_target:
                solved_flag_1st = 0
                # recover loading
                self.psse.read(0, step2_tempcase)  # saveRaw outputs .raw files only
                self.pfdt.getdata(self.psse)

                self.psse.scal_2(0, 1, 1, [0, 0, 0, 0, 0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
                self.psse.scal_2(
                    0,
                    1,
                    2,
                    [self._i, 1, 0, 1, 0],
                    [
                        self.pfd.load_mw_total,
                        self.pfdt.gen_mw_total / self.pfdt.load_mw_total * self.pfd.load_mw_total,
                        0.0,
                        -0.0,
                        0.0,
                        -0.0,
                        self._f,
                    ],
                )
                self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])

                solved_flag = self.if_solved(self.psse, 1, 5)

                if solved_flag:
                    logger.info("Should have solved the case.\n")
                    self.converged = True
                    return
                else:
                    self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, step2_tempcase)

                # recover dispatch
                if solved_flag == 0:
                    for i in range(len(self.pfd.gen_mw)):
                        busi = self.pfd.bus_num.index(self.pfd.gen_bus[i])

                        if self.pfd.gen_status[i] != 1 or abs(self.pfd.bus_type[busi]) != 2:
                            continue

                        self.psse.read(0, step2_tempcase)
                        self.psse.machine_chng_2(
                            self.pfd.gen_bus[i],
                            self.pfd.gen_id[i],
                            [self._i, self._i, self._i, self._i, self._i, self._i],
                            [
                                self.pfd.gen_mw[i],
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                                self._f,
                            ],
                        )
                        self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])
                        solved_flag = self.if_solved(self.psse, 1, 5)

                        if solved_flag == 0:
                            self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, step2_tempcase)  ## Possibility b
                        else:
                            logger.info("Should not see this!!!\n")
                            pass

            else:
                inc_cur = inc_cur - inc_step
                perc = str(inc_cur / inc_target * 100)[0:7]
                self.psse.read(0, step2_tempcase)

                self.save_raw(3, 2.1, perc)

                logger.info(
                    "PF (w added gens) not converge, but can converged at %" + perc + " loading. (Failed!)\n"
                )  ## Possibility c
                self.converged = False
                return

        ## IF-2, yes
        # if 1st power flow converges, try to remove added generators
        w_q_bus = []
        w_q_q = []
        if solved_flag_1st == 0:
            logger.info("PF (w added gens) converged.")
            step3_tempcase = self.save_raw(4, 3)

            # try to gradually remove all added generators
            # gen data from PSSE
            self.pfdt.getdata()
            ad_pq_genbus, ad_pq_gen_q, ad_pq_gen_q_abs, ad_pv_genbus, ad_pv_gen_q = self.get_added_generators(
                self.pfdt, ad_pq_busid, ad_pv_busid
            )

            logger.info("Trying to remove added generators..")
            # remove added gen
            n_ct_pv, n_ct_pq, idx = self.remove_added_generators(
                ad_pv_genbus, ad_pq_genbus, ad_pq_gen_q_abs, step3_tempcase
            )  ## Step 7

            # all added gens can be removed  ## IF-8
            if n_ct_pq == len(idx):  ## IF-8, yes
                logger.info("Orig PF converged at target loading. (Success!)\n")
                self.converged = True
                return

            ## IF-7, no
            # added gen cannot be removed completely
            # get Q at unremovable added gen
            ad_genbus, ad_gen_q = self.get_q_of_added_generator(step3_tempcase, ad_pq_busid, ad_pv_busid)

            # voltage adjustment to remove more added gen (directly solved by NR with a better initial)
            step4_tempcase = self.save_raw(5)
            q_remote, bus_remote, v_set_remote = self.remove_geni_adj_vi(ad_genbus, ad_gen_q, step4_tempcase)

            if len(bus_remote) == 0:
                logger.info("Orig PF converged (w adjusted local v_set). (Success!)\n")
                self.save_raw(1, 2)  ## Possibility e
                self.converged = True
                return

            # near-by PV bus voltage adjustment to remove even more added gen
            step5_tempcase = self.save_raw(5)
            self.pfdt.getdata()

            v_set = []
            v_bus = []
            for busi, qi, vseti in zip(bus_remote, q_remote, v_set_remote):
                nearpv, n_layer = self.get_near_pv(busi, 3)

                flag = 2
                for busj in nearpv:
                    gen_online = 0
                    last_found = -1
                    while busj in self.pfdt.gen_bus[last_found + 1 :]:
                        last_found = self.pfdt.gen_bus.index(busj, last_found + 1)
                        if last_found == -1:
                            break
                        else:
                            if self.pfdt.gen_status[last_found] == 1:
                                gen_online = 1

                    if gen_online == 1:
                        vj_set, flag_temp = self.remove_geni_adj_vj(busi, busj, step5_tempcase, vseti)
                    else:
                        continue

                    if flag_temp == 0:
                        flag = 0
                        v_bus.append(busj)
                        v_set.append(vj_set)
                        break

                if flag == 2:
                    w_q_bus.append(busi)
                    w_q_q.append(qi)

        if len(w_q_bus) == 0:
            logger.info("Orig PF converged (w adjusted remote v_set). (Success!)\n")
            self.save_raw(1, 2)  ## Possibility e
            self.converged = True
            return

        else:
            logger.info(
                "PF converged only when w Q support. "
                + str(len(w_q_bus))
                + " out of "
                + str(len(ad_pq_genbus) + len(ad_pv_genbus))
                + " left. \n"
            )
            self.save_raw(2)

            self.converged = True
            return

    @property
    def has_converged(self):
        return self.converged

    def save_raw(self, rawfile, option, option2=0, perc=0):
        # TODO: make changes here
        outfile = str(rawfile)
        if option == 1:
            outfile = outfile.replace("input", "dc2ac_output\\solved")
            if option2 == 1:
                outfile = outfile[0 : len(outfile) - 4] + """_solved.raw"""
            elif option2 == 2:
                outfile = outfile[0 : len(outfile) - 4] + """_addgen_remgen_solved.raw"""

        elif option == 2:
            outfile = outfile.replace("input", "dc2ac_output\\solvedwQ")
            outfile = outfile[0 : len(outfile) - 4] + """_step3_addgen_solved.raw"""

        elif option == 3:
            outfile = outfile.replace("input", "dc2ac_output\\unsolved")
            if option2 == 2:
                outfile = outfile[0 : len(outfile) - 4] + """_step2_addgen_unsolved.raw"""
            elif option2 == 2.1:
                outfile = outfile[0 : len(outfile) - 4] + """_step2_addgen_solved_""" + perc + """.raw"""

        elif option == 4:
            outfile = outfile.replace("input", "temp")
            if option2 == 1:
                outfile = outfile[0 : len(outfile) - 4] + """_step1_addgen.raw"""
            elif option2 == 2:
                outfile = outfile[0 : len(outfile) - 4] + """_step2_addgen_redP.raw"""
            elif option2 == 2.1:
                outfile = outfile[0 : len(outfile) - 4] + """_step2_addgen_solved.raw"""
            elif option2 == 3:
                outfile = outfile[0 : len(outfile) - 4] + """_step3_addgen_remgen.raw"""

        elif option == 5:
            outfile = outfile.replace("input", "dc2ac_output\\solved_Vadju")
            outfile = outfile[0 : len(outfile) - 4] + """_step4_unlimQ.raw"""

        self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, outfile)

        return outfile

    def if_solved(self, option, n_flag1_max):
        if option == 1:
            self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])
        else:
            self.psse.fnsl([0, 0, 0, 1, 1, 1, 0])  # NR with flat start
        solved_flag = self.psse.solved()

        n_flag1 = 0
        while solved_flag == 1:
            n_flag1 = n_flag1 + 1
            if option == 1:
                self.psse.nsol([0, 0, 0, 1, 1, 0, 0])  # fast decouple with "Do not flat start"
            else:
                self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])
            solved_flag = self.psse.solved()
            if n_flag1 > n_flag1_max:
                logger.info("Flag = 1 does not disappear over " + str(n_flag1_max) + " runs. Let Flag = 2.")
                solved_flag = 2
        return solved_flag

    def add_generators(self):
        ad_pq_busid = []
        ad_pv_busid = []
        for ii in self.all_subs:
            bus_i = self.pfd.bus_num.index(ii)
            bus_i_type = self.pfd.bus_type[bus_i]

            if bus_i_type == 1:
                ad_pq_busid.append(ii)
                self.psse.bus_chng_3(
                    ii,
                    [2, self._i, self._i, self._i],
                    [self._f, 1.000, self._f, self._f, self._f, self._f, self._f],
                    self._s,
                )
                self.psse.plant_data(ii, self._i, [1.000, self._f])
                self.psse.machine_data_2(
                    ii,
                    r"""ad""",
                    [self._i, self._i, self._i, self._i, self._i, self._i],
                    [
                        self._f,
                        self._f,
                        99999.0,
                        -99999.0,
                        99999.0,
                        -99999.0,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                    ],
                )
            if bus_i_type == 2:
                ad_pv_busid.append(ii)
                self.psse.bus_chng_3(
                    ii,
                    [2, self._i, self._i, self._i],
                    [self._f, self._f, self._f, self._f, self._f, self._f, self._f],
                    self._s,
                )
                self.psse.machine_data_2(
                    ii,
                    r"""ad""",
                    [self._i, self._i, self._i, self._i, self._i, self._i],
                    [
                        self._f,
                        self._f,
                        99999.0,
                        -99999.0,
                        99999.0,
                        -99999.0,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                    ],
                )
        return (ad_pq_busid, ad_pv_busid)

    def svblt_check(self):
        # step 1: get a solvable case by reducing the loading
        dec_perc0 = float(0)
        dec_inc = float(5)
        dec_perc = dec_perc0
        n_iter = 0
        n_iter_max = 4
        solved_flag = 1

        insolvable_flag = 0
        while (solved_flag != 0) & (n_iter < n_iter_max):
            n_iter = n_iter + 1
            dec_perc = dec_perc + dec_inc

            # early termination
            if dec_perc >= 10:
                insolvable_flag = 1
                break

            new_load_mw = (100 - dec_perc) / 100 * self.pfd.load_mw_total
            nnew_load_mvar = (100 - dec_perc) / 100 * self.pfd.load_Mvar_total
            new_gen_mw = (100 - dec_perc) / 100 * self.pfd.gen_mw_total

            self.solver.reload()

            self.psse.scal_2(0, 1, 1, [0, 0, 0, 0, 0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
            self.psse.scal_2(
                0, 1, 2, [self.psse._i, 1, 0, 1, 0], [new_load_mw, new_gen_mw, 0.0, -0.0, 0.0, -0.0, nnew_load_mvar]
            )

            self.psse.fnsl([0, 0, 0, 1, 1, 1, 99, 0])
            solved_flag = self.psse.solved()
        if insolvable_flag == 1:
            logger.info("Power flow cannot converge above " + str(100 - dec_perc) + "% loading.\n")
        return insolvable_flag, dec_perc

    def approach_target_loading(self, dec_perc, step2_tempcase):
        inc_target = 100 / (100 - float(dec_perc))
        inc_cur = 1.0001
        inc_step = 0.01
        inc_step_mw = self.pfd.gen_mw_total / inc_target * inc_cur * inc_step
        inc_step_mw_min = float(1)

        pfdt = PowerFlowData(self.psse)

        self.psse.read(0, step2_tempcase)
        pfdt.getdata(self.psse)

        for i in range(len(pfdt.gen_bus)):
            busi = pfdt.bus_num.index(pfdt.gen_bus[i])
            if pfdt.bus_type[busi] == 3 and pfdt.gen_status[i] == 1:
                pfdt.gen_mw_total = pfdt.gen_mw_total - pfdt.gen_mw[i]

        load_mw_pre = pfdt.load_mw_total
        gen_mw_pre = pfdt.gen_mw_total

        while (inc_cur < inc_target) & (inc_step_mw > inc_step_mw_min):
            inc_cur = inc_cur + inc_step
            self.psse.read(0, step2_tempcase)

            gen_mw = gen_mw_pre * inc_cur
            load_mw = load_mw_pre * inc_cur

            for i in range(len(pfdt.gen_bus)):
                busi = pfdt.bus_num.index(pfdt.gen_bus[i])
                if pfdt.bus_type[busi] == 3 and pfdt.gen_status[i] == 1:
                    self.psse.machine_chng_2(
                        pfdt.gen_bus[i],
                        pfdt.gen_id[i],
                        [self._i, self._i, self._i, self._i, self._i, self._i],
                        [
                            0.0,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                            self._f,
                        ],
                    )
            self.psse.scal_2(0, 1, 1, [0, 0, 0, 0, 0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
            self.psse.scal_2(0, 1, 2, [self._i, 1, 0, 1, 0], [load_mw, gen_mw, 0.0, -0.0, 0.0, -0.0, self._f])

            solved_flag = self.if_solved(1, 5)
            if solved_flag == 0:
                self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, step2_tempcase)
            else:
                inc_cur = inc_cur - inc_step
                inc_step = inc_step / 2
            inc_step_mw = load_mw_pre * inc_step

        return inc_cur, inc_target, inc_step

    def remove_added_generators(self, ad_pv_genbus, ad_pq_genbus, ad_pq_gen_q_abs, step3_tempcase):
        # remove added gen on original pv bus
        n_ct_pv = 0
        for ii in ad_pv_genbus:
            self.psse.read(0, step3_tempcase)
            self.psse.purgmac(ii, r"""ad""")

            solved_flag = self.ifsolved(self.psse, 1, 3)

            if solved_flag == 0:
                n_ct_pv = n_ct_pv + 1
                self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, step3_tempcase)
            if solved_flag == 2:
                break

        # remove added gen on original pq bus
        idx = sorted(range(len(ad_pq_gen_q_abs)), key=lambda k: ad_pq_gen_q_abs[k])
        n_ct_pq = 0
        n_step = 32
        n_cur = 0
        n_max = len(idx)
        while True:
            if n_step == 0:
                break
            if n_cur + n_step > n_max:
                n_step = n_max - n_cur

            self.psse.read(0, step3_tempcase)
            for nn in range(n_cur, n_cur + n_step, 1):
                ii = idx[nn]
                self.psse.purgmac(ad_pq_genbus[ii], r"""ad""")
                self.psse.bus_chng_3(
                    ad_pq_genbus[ii],
                    [1, self._i, self._i, self._i],
                    [self._f, self._f, self._f, self._f, self._f, self._f, self._f],
                    self._s,
                )

            solved_flag = self.ifsolved(self.psse, 1, 3)

            if solved_flag == 0:
                n_ct_pq = n_ct_pq + n_step
                n_cur = n_cur + n_step
                self.psse.rawd_2(0, 1, [1, 1, 1, 0, 0, 0, 0], 0, step3_tempcase)
            else:
                if n_step >= 1:
                    if n_step == 2:
                        n_step = 1
                    else:
                        n_step = int(n_step / 4)
                else:
                    break
        return n_ct_pv, n_ct_pq, idx

    def sol_out_orig(self, solved_flag):
        if solved_flag == 0:
            logger.info("Orig PF converged at target loading. (Success!)\n")
        else:
            logger.info("Orig PF cannot converge at target loading.")

    def get_bus_list(self):
        bus_list = []
        for bus in self.raw_data.buses:
            ierr, kv = self.psse.busdat(bus, "KV")
            assert ierr == 0, f"error={ierr}"
            if kv > self.min_kv_filter and random.random() < self.filter_buses:
                bus_list.append(bus)

        return bus_list

    def get_added_generators(self, pfdt, ad_pq_busid, ad_pv_busid):
        ad_pq_genbus = []
        ad_pq_gen_q = []
        ad_pq_gen_q_abs = []

        ad_pv_genbus = []
        ad_pv_gen_q = []

        for ii in range(len(pfdt.gen_bus)):
            if pfdt.gen_id[ii] == "AD":
                if pfdt.gen_bus[ii] in ad_pq_busid:
                    ad_pq_genbus.append(pfdt.gen_bus[ii])
                    ad_pq_gen_q.append(pfdt.gen_Mvar[ii])
                    ad_pq_gen_q_abs.append(abs(pfdt.gen_Mvar[ii]))
                if pfdt.gen_bus[ii] in ad_pv_busid:
                    ad_pv_genbus.append(pfdt.gen_bus[ii])
                    ad_pv_gen_q.append(pfdt.gen_Mvar[ii])
        return ad_pq_genbus, ad_pq_gen_q, ad_pq_gen_q_abs, ad_pv_genbus, ad_pv_gen_q

    def get_q_of_added_generator(self, step3_tempcase, ad_pq_busid, ad_pv_busid):
        self.psse.read(0, step3_tempcase)
        self.psse.fnsl([0, 0, 0, 1, 1, 0, -1, 0])
        temp_genid = self.psse.amachchar(-1, 4, "ID")[1][0]
        temp_genbus = self.psse.amachint(-1, 4, "NUMBER")[1][0]
        temp_genoutput = np.asarray(self.psse.amachcplx(-1, 4, "PQGEN")[1][0])
        # temp_gen_mw = temp_genoutput.real
        temp_gen_mvar = temp_genoutput.imag

        ad_genbus = []
        ad_gen_q = []
        for ii in range(len(temp_genbus)):
            if temp_genid[ii] == "AD":
                if temp_genbus[ii] in ad_pq_busid:
                    ad_genbus.append(temp_genbus[ii])
                    ad_gen_q.append(temp_gen_mvar[ii])
                if temp_genbus[ii] in ad_pv_busid:
                    ad_genbus.append(temp_genbus[ii])
                    ad_gen_q.append(temp_gen_mvar[ii])
        return ad_genbus, ad_gen_q

    # By adjusting V set point at PV bus, those with close-to-zero
    # qinj can be removed directly (due to a better initial condition for NR)
    def remove_geni_adj_vj(self, ad_genbus, ad_gen_q, step4_tempcase, delta_thrd=1000):
        bus_remote = []
        q_remote = []
        v_set_remote = []
        for busi, q_i in zip(ad_genbus, ad_gen_q):
            q = []
            self.psse.read(0, step4_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [1.00, self._f])
            flag = self.ifsolved(self.psse, 1, 5)
            self.pfdt.getdata(self.psse)
            idxi = self.pfdt.gen_bus.index(busi)
            while self.pfdt.gen_id[idxi] != "AD":
                idxi = idxi + 1
            q.append(self.pfdt.gen_Mvar[idxi])

            self.psse.read(0, step4_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [0.99, self._f])
            flag = self.ifsolved(self.psse, 1, 5)
            self.pfdt.getdata(self.psse)
            idxi = self.pfdt.gen_bus.index(busi)
            while self.pfdt.gen_id[idxi] != "AD":
                idxi = idxi + 1
            q.append(self.pfdt.gen_Mvar[idxi])

            self.psse.read(0, step4_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [1.01, self._f])
            flag = self.ifsolved(self.psse, 1, 5)
            self.pfdt.getdata(self.psse)
            idxi = self.pfdt.gen_bus.index(busi)
            while self.pfdt.gen_id[idxi] != "AD":
                idxi = idxi + 1
            q.append(self.pfdt.gen_Mvar[idxi])

            a = (-1.0000 * q[0] + 0.5000 * q[1] + 0.5000 * q[2]) * 10000
            b = (2.0000 * q[0] - 1.0050 * q[1] - 0.9950 * q[2]) * 10000
            c = (-0.9999 * q[0] + 0.5050 * q[1] + 0.4950 * q[2]) * 10000

            delta = b * b - 4 * a * c

            if delta > delta_thrd:
                x1 = (-b + np.sqrt(b * b - 4 * a * c)) / 2 / a
                x2 = (-b - np.sqrt(b * b - 4 * a * c)) / 2 / a
                if abs(x1 - 1) < abs(x2 - 1):
                    vset = x1
                else:
                    vset = x2
                self.psse.read(0, step4_tempcase)
                self.psse.plant_chng_4(busi, 0, [self._i, self._i], [vset, self._f])
                self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])
                flag = self.ifsolved(self.psse, 1, 5)

                self.psse.purgmac(busi, r"""ad""")
                self.psse.bus_chng_3(
                    busi,
                    [1, self._i, self._i, self._i],
                    [self._f, self._f, self._f, self._f, self._f, self._f, self._f],
                    self._s,
                )
                flag = self.ifsolved(self.psse, 1, 3)
            else:
                flag = 2

            if flag == 0:
                pass
                # step4_tempcase = saveRaw(self.psse, rawfile, 5)
            else:
                logger.info("cannot remove gen" + str(busi))
                self.psse.read(0, step4_tempcase)

                self.pfdt.getdata(self.psse)
                idxi = self.pfdt.gen_bus.index(busi)
                while self.pfdt.gen_id[idxi] != "AD":
                    idxi = idxi + 1
                q_v_set = self.pfdt.gen_Mvar[idxi]

                bus_remote.append(busi)
                q_remote.append(q_v_set)
                v_set_remote.append(vset)
        return q_remote, bus_remote, v_set_remote

    # get nearby buses - single layer
    def get_near_by_bus(self, busi):
        nearbus = []
        nearpv = []
        last_found = -1
        while busi in self.pfdt.brc_from[last_found + 1 :]:
            last_found = self.pfdt.brc_from.index(busi, last_found + 1)
            if last_found == -1:
                break
            else:
                if (self.pfdt.brc_to[last_found] not in nearbus) & (self.pfdt.brc_to[last_found] != busi):
                    nearbus.append(self.pfdt.brc_to[last_found])
                    idx = self.pfdt.bus_num.index(self.pfdt.brc_to[last_found])
                    if self.pfdt.bus_type[idx] == 2:
                        nearpv.append(self.pfdt.brc_to[last_found])
        return nearbus, nearpv

    # get nearby pv buses
    def get_near_pv(self, busi, n_layer=1):
        # get nearby buses and nearby pv buses
        n = 1
        nearbus, nearpv = self.get_near_by_bus(busi)

        if len(nearpv) == 0:
            n_layer = n_layer + 1

        while n < n_layer:
            n = n + 1
            nearbus_pre = []
            for busii in nearbus:
                nearbus_pre.append(busii)

            for busii in nearbus_pre:
                nearbus_temp, nearpv_temp = self.get_near_by_bus(busii)
                for ii in range(len(nearbus_temp)):
                    if (nearbus_temp[ii] not in nearbus) & (nearbus_temp[ii] != busi):
                        nearbus.append(nearbus_temp[ii])
                for ii in range(len(nearpv_temp)):
                    if (nearpv_temp[ii] not in nearpv) & (nearpv_temp[ii] != busi):
                        nearpv.append(nearpv_temp[ii])
            if len(nearpv) == 0:
                n_layer = n_layer + 1

        return nearpv, n_layer

    def remove_geni_adj_vj(self, busi, busj, pfdt, step5_tempcase, vseti):
        q = []
        self.psse.read(0, step5_tempcase)
        self.psse.plant_chng_4(busi, 0, [self._i, self._i], [vseti, self._f])
        self.psse.plant_chng_4(busj, 0, [self._i, self._i], [1.0, self._f])
        flag = self.ifsolved(self.psse, 1, 5)
        pfdt.getdata(self.psse)
        idxi = pfdt.gen_bus.index(busi)
        while pfdt.gen_id[idxi] != "AD":
            idxi = idxi + 1
        q.append(pfdt.gen_Mvar[idxi])

        self.psse.read(0, step5_tempcase)
        self.psse.plant_chng_4(busi, 0, [self._i, self._i], [vseti, self._f])
        self.psse.plant_chng_4(busj, 0, [self._i, self._i], [1.01, self._f])
        flag = self.ifsolved(self.psse, 1, 5)
        pfdt.getdata(self.psse)
        idxi = pfdt.gen_bus.index(busi)
        while pfdt.gen_id[idxi] != "AD":
            idxi = idxi + 1
        q.append(pfdt.gen_Mvar[idxi])

        a = (-1.0000 * q[0] + 1.0000 * q[1]) * 100
        b = q[0] - a
        if a != 0:
            vset = -b / a + 0.01

            self.psse.read(0, step5_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [vseti, self._f])
            self.psse.plant_chng_4(busj, 0, [self._i, self._i], [vset, self._f])
            self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])
            flag = self.ifsolved(self.psse, 1, 5)

            self.psse.purgmac(busi, r"""ad""")
            self.psse.bus_chng_3(
                busi,
                [1, self._i, self._i, self._i],
                [self._f, self._f, self._f, self._f, self._f, self._f, self._f],
                self._s,
            )
            flag = self.ifsolved(self.psse, 1, 3)
        else:
            vset = 0
            flag = 2

        return vset, flag

    def remove_geni_adj_vi(self, ad_genbus, ad_gen_q, step4_tempcase, delta_thrd=1000):
        bus_remote = []
        q_remote = []
        v_set_remote = []
        for busi, q_i in zip(ad_genbus, ad_gen_q):
            q = []
            self.psse.read(0, step4_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [1.00, self._f])
            flag = self.if_solved(1, 5)
            self.pfdt.getdata(self.psse)
            idxi = self.pfdt.gen_bus.index(busi)
            while self.pfdt.gen_id[idxi] != "AD":
                idxi = idxi + 1
            q.append(self.pfdt.gen_Mvar[idxi])

            self.psse.read(0, step4_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [0.99, self._f])
            flag = self.if_solved(1, 5)
            self.pfdt.getdata(self.psse)
            idxi = self.pfdt.gen_bus.index(busi)
            while self.pfdt.gen_id[idxi] != "AD":
                idxi = idxi + 1
            q.append(self.pfdt.gen_Mvar[idxi])

            self.psse.read(0, step4_tempcase)
            self.psse.plant_chng_4(busi, 0, [self._i, self._i], [1.01, self._f])
            flag = self.if_solved(1, 5)
            self.pfdt.getdata(self.psse)
            idxi = self.pfdt.gen_bus.index(busi)
            while self.pfdt.gen_id[idxi] != "AD":
                idxi = idxi + 1
            q.append(self.pfdt.gen_Mvar[idxi])

            a = (-1.0000 * q[0] + 0.5000 * q[1] + 0.5000 * q[2]) * 10000
            b = (2.0000 * q[0] - 1.0050 * q[1] - 0.9950 * q[2]) * 10000
            c = (-0.9999 * q[0] + 0.5050 * q[1] + 0.4950 * q[2]) * 10000

            delta = b * b - 4 * a * c

            if delta > delta_thrd:
                x1 = (-b + np.sqrt(b * b - 4 * a * c)) / 2 / a
                x2 = (-b - np.sqrt(b * b - 4 * a * c)) / 2 / a
                if abs(x1 - 1) < abs(x2 - 1):
                    vset = x1
                else:
                    vset = x2
                self.psse.read(0, step4_tempcase)
                self.psse.plant_chng_4(busi, 0, [self._i, self._i], [vset, self._f])
                self.psse.fnsl([0, 0, 0, 1, 1, 0, 0, 0])
                flag = self.if_solved(1, 5)

                self.psse.purgmac(busi, r"""ad""")
                self.psse.bus_chng_3(
                    busi,
                    [1, self._i, self._i, self._i],
                    [self._f, self._f, self._f, self._f, self._f, self._f, self._f],
                    self._s,
                )
                flag = self.if_solved(1, 3)
            else:
                flag = 2

            if flag == 0:
                pass
                # step4_tempcase = self.save_raw(self.psse, rawfile, 5)
            else:
                logger.info("cannot remove gen" + str(busi))
                self.psse.read(0, step4_tempcase)

                self.pfdt.getdata(self.psse)
                idxi = self.pfdt.gen_bus.index(busi)
                while self.pfdt.gen_id[idxi] != "AD":
                    idxi = idxi + 1
                q_v_set = self.pfdt.gen_Mvar[idxi]

                bus_remote.append(busi)
                q_remote.append(q_v_set)
                v_set_remote.append(vset)
        return q_remote, bus_remote, v_set_remote



from pypsse.data_writers.csv import CSVWriter
from pypsse.data_writers.hdf5 import HDF5Writer
from pypsse.data_writers.json import JSONWriter


class DummyWriter:
    def __init__(self, *_, **__):
        return

    def write(self, *_, **__):
        return


class DataWriter:
    "Data writer class defination"
    modes = {
        "h5": HDF5Writer,
        "csv": CSVWriter,
        "json": JSONWriter,
        "none": DummyWriter,
    }

    def __init__(self, log_dir, formatnm, column_length):
        "Sets up a data writer as per user input"
        self.writer = self.modes[formatnm](log_dir, column_length)

    def write(self, currenttime, powerflow_output, convergence):
        "Enables incremental write to the data writer object"
        self.writer.write(currenttime, powerflow_output, convergence)



from typing import Union

from loguru import logger

from pypsse.models import ExportFileOptions, SimulationSettings
from pypsse.modes.dynamic import Dynamic
from pypsse.modes.pcm import ProductionCostModel
from pypsse.modes.snap import Snap
from pypsse.modes.static import Static
from pypsse.parsers.reader import Reader


def sim_controller(
    psse: object,
    dyntools: object,
    settings: SimulationSettings,
    export_settings: ExportFileOptions,
    subsystem_buses: dict,
    raw_data: Reader,
) -> Union[Dynamic, ProductionCostModel, Snap, Static]:
    """sets up an appropriate simualtion controller based on user input

    Args:
        psse (object): simulator instance
        dyntools (object): psse dyntools instance
        settings (SimulationSettings): simulation settings
        export_settings (ExportFileOptions): export settings
        subsystem_buses (dict): mapping of bus subsystems to buses
        raw_data (Reader): instance of model reader

    Returns:
        Union[Dynamic, ProductionCostModel, Snap, Static]: simulator controller instance
    """

    sim_modes = {"Dynamic": Dynamic, "Steady-state": Static, "Snap": Snap, "ProductionCostModel": ProductionCostModel}

    sim = sim_modes[settings.simulation.simulation_mode.value](
        psse, dyntools, settings, export_settings, subsystem_buses, raw_data
    )
    logger.debug(f"Simulator contoller of type {settings.simulation.simulation_mode.value} created")
    return sim



import os
from typing import ClassVar, List

import pandas as pd
from loguru import logger

from pypsse.common import MACHINE_CHANNELS
from pypsse.modes.constants import dyn_only_options


class DynamicUtils:
    "Utility functions for dynamic simulations"

    dynamic_params: ClassVar[List[str]] = ["FmA", "FmB", "FmC", "FmD", "Fel"]

    def disable_generation_for_coupled_buses(self):
        """Disables generation of coupled buses (co-simulation mode only)"""
        if (
            self.settings.helics
            and self.settings.helics.cosimulation_mode
            and self.settings.helics.disable_generation_on_coupled_buses
        ):
            sub_data = pd.read_csv(self.settings.simulation.subscriptions_file)
            sub_data = sub_data[sub_data["element_type"] == "Load"]
            generators = {}
            generator_list = {}

            for gen_bus, gen_id in self.raw_data.generators:
                if gen_bus not in generator_list:
                    generator_list[gen_bus] = []
                generator_list[gen_bus].append(gen_id)

            for _, row in sub_data.iterrows():
                bus = row["bus"]
                generators[bus] = generator_list[bus]

            for bus_id, machines in generators.items():
                for machine in machines:
                    intgar = [0, self._i, self._i, self._i, self._i, self._i]
                    realar = [
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                        self._f,
                    ]
                    self.psse.machine_chng_2(bus_id, machine, intgar, realar)
                    logger.info(f"Machine disabled: {bus_id}_{machine}")

    def disable_load_models_for_coupled_buses(self):
        """Disables loads of coupled buses (co-simulation mode only)"""
        if self.settings.helics and self.settings.helics.cosimulation_mode:
            sub_data = pd.read_csv(self.settings.simulation.subscriptions_file)
            sub_data = sub_data[sub_data["element_type"] == "Load"]

            self.psse_dict = {}
            for _, row in sub_data.iterrows():
                bus = row["bus"]
                load = row["element_id"]
                ierr = self.psse.ldmod_status(0, int(bus), str(load), 1, 0)
                assert ierr == 0, f"error={ierr}"
                logger.error(f"Dynamic model for load {load} connected to bus {bus} has been disabled")

    def break_loads(self, loads: list = None, components_to_replace: List[str] = []):
        """Implements the load split logic

        Args:
            loads (list, optional): list of coupled loads. Defaults to None.
            components_to_replace (List[str], optional): components to be simulated on distribution side. Defaults to [].
        """

        components_to_stay = [x for x in self.dynamic_params if x not in components_to_replace]
        if loads is None:
            loads = self._get_coupled_loads()
        loads = self._get_load_static_data(loads)
        loads = self._get_load_dynamic_data(loads)
        loads = self._replicate_coupled_load(loads, components_to_replace)
        self._update_dynamic_parameters(loads, components_to_stay, components_to_replace)

    def _update_dynamic_parameters(self, loads: dict, components_to_stay: list, components_to_replace: list):
        """Updates dynamic parameters of composite old / replicated load models

        Args:
            loads (dict): load dictionary
            components_to_stay (list): components to be simulated on transmission side
            components_to_replace (list): components to be simulated on distribution side
        """

        new_percentages = {}
        for load in loads:
            count = 0
            for comp in components_to_stay:
                count += load[comp]
            for comp in components_to_stay:
                new_percentages[comp] = load[comp] / count
            for comp in components_to_replace:
                new_percentages[comp] = 0.0

            settings = self._get_load_dynamic_properties(load)
            #
            for k, v in new_percentages.items():
                idx = dyn_only_options["Loads"]["lmodind"][k]
                settings[idx] = v
                # self.psse.change_ldmod_con(load['bus'], 'XX' ,r"""CMLDBLU2""" ,idx ,v)
            values = list(settings.values())
            self.psse.add_load_model(load["bus"], "XX", 0, 1, r"""CMLDBLU2""", 2, [0, 0], ["", ""], 133, values)
            logger.info(f"Dynamic model parameters for load {load['id']} at bus 'XX' changed.")

    def _get_load_dynamic_properties(self, load):
        "Returns dynamic parameters of composite load models"
        settings = {}
        for i in range(133):
            ierr, con_index = self.psse.lmodind(load["bus"], str(load["id"]), "CHARAC", "CON")
            if con_index is not None:
                act_con_index = con_index + i
                ierr, value = self.psse.dsrval("CON", act_con_index)
                assert ierr == 0, f"error={ierr}"
                settings[i] = value
        return settings

    def _replicate_coupled_load(self, loads: dict, components_to_replace: list):
        """create a replica of composite load model

        Args:
            loads (dict): load dictionary
            components_to_replace (list): composite load models to replace on distribution side

        Returns:
            dict: updated load dictionary
        """

        for load in loads:
            dynamic_percentage = load["FmA"] + load["FmB"] + load["FmC"] + load["FmD"] + load["Fel"]
            static_percentage = 1.0 - dynamic_percentage
            for comp in components_to_replace:
                static_percentage += load[comp]
            remaining_load = 1 - static_percentage
            total_load = load["MVA"]
            total_distribution_load = total_load * static_percentage
            total_transmission_load = total_load * remaining_load
            # ceate new load
            self.psse.load_data_5(
                load["bus"],
                "XX",
                realar=[total_transmission_load.real, total_transmission_load.imag, 0.0, 0.0, 0.0, 0.0],
                lodtyp="replica",
            )
            # ierr, cmpval = self.psse.loddt2(load["bus"], "XX" ,"MVA" , "ACT")
            # modify old load
            self.psse.load_data_5(
                load["bus"],
                str(load["id"]),
                realar=[total_distribution_load.real, total_distribution_load.imag, 0.0, 0.0, 0.0, 0.0],
                lodtyp="original",
            )
            # ierr, cmpval = self.psse.loddt2(load["bus"], load["id"] ,"MVA" , "ACT")
            logger.info(f"Original load {load['id']} @ bus {load['bus']}: {total_load}")
            logger.info(f"New load 'XX' @ bus {load['bus']} created successfully: {total_transmission_load}")
            logger.info(f"Load {load['id']} @ bus {load['bus']} updated : {total_distribution_load}")
            load["distribution"] = total_distribution_load
            load["transmission"] = total_transmission_load
        return loads

    def _get_coupled_loads(self) -> list:
        """Returns a list of all coupled loads ina give simualtion

        Returns:
            list: list of coupled loads
        """

        sub_data = pd.read_csv(
            os.path.join(
                self.settings["Simulation"]["Project Path"], "Settings", self.settings["HELICS"]["Subscriptions file"]
            )
        )
        load = []
        for _, row in sub_data.iterrows():
            if row["element_type"] == "Load":
                load.append(
                    {
                        "type": row["element_type"],
                        "id": row["element_id"],
                        "bus": row["bus"],
                    }
                )
        return load

    def _get_load_static_data(self, loads: list) -> dict:
        """Returns static data for load models

        Args:
            loads (list): list of load names

        Returns:
            dict: mapping load to static values
        """

        values = ["MVA", "IL", "YL", "TOTAL"]
        for load in loads:
            for v in values:
                ierr, cmpval = self.psse.loddt2(load["bus"], str(load["id"]), v, "ACT")
                load[v] = cmpval
        return loads

    def _get_load_dynamic_data(self, loads: list) -> dict:
        """Returns dynamic data for load models

        Args:
            loads (list): list of load names

        Returns:
            dict: mapping load to dynamic values
        """

        values = dyn_only_options["Loads"]["lmodind"]
        for load in loads:
            for v, con_ind in values.items():
                ierr = self.psse.inilod(load["bus"])
                assert ierr == 0, f"error={ierr}"
                ierr, ld_id = self.psse.nxtlod(load["bus"])
                assert ierr == 0, f"error={ierr}"
                if ld_id is not None:
                    ierr, con_index = self.psse.lmodind(load["bus"], ld_id, "CHARAC", "CON")
                    assert ierr == 0, f"error={ierr}"
                    if con_index is not None:
                        act_con_index = con_index + con_ind
                        ierr, value = self.psse.dsrval("CON", act_con_index)
                        assert ierr == 0, f"error={ierr}"
                        load[v] = value
        return loads

    def setup_machine_channels(self, machines: dict, properties: list):
        """sets up machine channels

        Args:
            machines (dict): mapping machine to connected bus
            properties (list): list of machine properties
        """

        for _, qty in enumerate(properties):
            if qty not in self.channel_map:
                nqty = f"MACHINE_{qty}"
                self.channel_map[nqty] = {}
            for mch, b in machines:
                if qty in MACHINE_CHANNELS:
                    self.channel_map[nqty][f"{b}_{mch}"] = [self.chnl_idx]
                    chnl_id = MACHINE_CHANNELS[qty]
                    logger.info(f"{qty} for machine {b}_{mch} added to channel {self.chnl_idx}")
                    self.psse.machine_array_channel([self.chnl_idx, chnl_id, int(b)], mch, "")
                    self.chnl_idx += 1

    def setup_load_channels(self, loads: list):
        """Sets up load channels

        Args:
            loads (list): list of loads
        """

        if "LOAD_P" not in self.channel_map:
            self.channel_map["LOAD_P"] = {}
            self.channel_map["LOAD_Q"] = {}

        for ld, b in loads:
            self.channel_map["LOAD_P"][f"{b}_{ld}"] = [self.chnl_idx]
            self.channel_map["LOAD_Q"][f"{b}_{ld}"] = [self.chnl_idx + 1]
            self.psse.load_array_channel([self.chnl_idx, 1, int(b)], ld, "")
            self.psse.load_array_channel([self.chnl_idx + 1, 2, int(b)], ld, "")
            logger.info(f"P and Q for load {b}_{ld} added to channel {self.chnl_idx} and {self.chnl_idx + 1}")
            self.chnl_idx += 2

    def setup_bus_channels(self, buses: list, properties: list):
        """Sets up bus channels

        Args:
            buses (list): list of buses
            properties (dict): list of bus properties
        """

        for _, qty in enumerate(properties):
            if qty not in self.channel_map:
                self.channel_map[qty] = {}
            for _, b in enumerate(buses):
                if qty == "frequency":
                    self.channel_map[qty][b] = [self.chnl_idx]
                    self.psse.bus_frequency_channel([self.chnl_idx, int(b)], "")
                    logger.info(f"Frequency for bus {b} added to channel { self.chnl_idx}")
                    self.chnl_idx += 1
                elif qty == "voltage_and_angle":
                    self.channel_map[qty][b] = [self.chnl_idx, self.chnl_idx + 1]
                    self.psse.voltage_and_angle_channel([self.chnl_idx, -1, -1, int(b)], "")
                    logger.info(f"Voltage and angle for bus {b} added to channel {self.chnl_idx} and {self.chnl_idx+1}")
                    self.chnl_idx += 2

    def poll_channels(self) -> dict:
        """Polls all channels adde during the setup process

        Returns:
            dict: mapping of polled channels to values
        """

        results = {}
        for ppty, b_dict in self.channel_map.items():
            ppty_new = ppty.split("_and_")
            for b, indices in b_dict.items():
                for n, idx in zip(ppty_new, indices):
                    if "_" not in n:
                        n_name = f"BUS_{n}"
                    else:
                        n_name = n
                    if n_name not in results:
                        results[n_name] = {}
                    ierr, value = self.psse.chnval(idx)
                    assert ierr == 0, f"error={ierr}"
                    if value is None:
                        value = -1
                    results[n_name][b] = value
        return results

    def setup_all_channels(self):
        """Sets up all user-defined channels for a project"""

        self.channel_map = {}
        self.chnl_idx = 1
        if not self.export_settings.channel_setup:
            return

        for channel in self.export_settings.channel_setup:
            method_type = channel.asset_type
            if method_type == "buses":
                self.setup_bus_channels(channel.asset_list, channel.asset_properties)
            elif method_type == "loads":
                load_list = [[x, int(y)] for x, y in channel.asset_list]
                self.setup_load_channels(load_list)
            elif method_type == "machines":
                machine_list = [[x, int(y)] for x, y in channel.asset_list]
                self.setup_machine_channels(machine_list, channel.asset_properties)



"""
@author:Aadil Latif
@file:pyPSSE.py
@time:2/4/2020
"""

import json
import os
import subprocess
import sys
import time
from pathlib import Path
from typing import List, Union

import numpy as np
import pandas as pd
import toml
from loguru import logger
from networkx import Graph

import pypsse.contingencies as c
import pypsse.simulation_controller as sc
from pypsse.common import (
    EXPORTS_SETTINGS_FILENAME,
    LOGS_FOLDER,
    MAX_PSSE_BUSSYSTEMS,
    SIMULATION_SETTINGS_FILENAME,
)
from pypsse.contingencies import (
    BusFaultObject,
    BusTripObject,
    LineFaultObject,
    LineTripObject,
    MachineTripObject,
)
from pypsse.enumerations import SimulationStatus
from pypsse.helics_interface import HelicsInterface
from pypsse.models import (
    BusSubsystems,
    ExportAssetTypes,
    ExportFileOptions,
    SimulationModes,
    SimulationSettings,
)
from pypsse.parsers import gic_parser as gp
from pypsse.parsers import reader as rd
from pypsse.profile_manager.profile_store import ProfileManager
from pypsse.result_container import Container
from pypsse.models import Contingencies
from pypsse.enumerations import PSSE_VERSIONS

USING_NAERM = 0

N_BUS = 200000


class Simulator:
    "Base class for the simulator"

    _status: SimulationStatus = SimulationStatus.NOT_INITIALIZED

    def __init__(
        self,
        settings: SimulationSettings,
        export_settings: Union[ExportFileOptions, None] = None,
        psse_version: str = PSSE_VERSIONS.PSSE35.value,
    ):
        """Load a valid PyPSSE project and sets up simulation

        Args:
            settings (SimulationSettings): simulation settings
            export_settings (Union[ExportFileOptions, None]): export settings
            psse_path (Union[str, Path], optional): Path to python environment within the PSS/e install directory
        """

        self._status = SimulationStatus.STARTING_INSTANCE
        self.settings = settings

        logger.debug(f"Instantiating psse version {psse_version}")
        __import__(psse_version, fromlist=[""])  # noqa: F401

        import dyntools
        import psspy

        ierr = psspy.psseinit(N_BUS)
        assert ierr == 0, f"Error code: {ierr}"

        if export_settings is None:
            export_settings_path = (
                Path(self.settings.simulation.project_path)
                / EXPORTS_SETTINGS_FILENAME
            )
            assert (
                export_settings_path.exists()
            ), f"{export_settings_path} does nor exist"
            export_settings = toml.load(export_settings_path)
            export_settings = ExportFileOptions(**export_settings)

        self.export_settings = export_settings
        log_path = os.path.join(
            self.settings.simulation.project_path, LOGS_FOLDER
        )
        logger.debug("Starting PSSE instance")

        self.dyntools = dyntools
        self.psse = psspy
        # logger.debug('Initializing PSS/E. connecting to license server')

        self.start_simulation()
        self.init()
        self._status = SimulationStatus.INITIALIZATION_COMPLETE

    @classmethod
    def from_setting_files(
        cls, simulation_settiings_file: Path, export_Settings_file: Path = None
    ):
        """build 'Simulator' from toml settings files

        Args:
            simulation_settiings_file (Path): simulation settings
            export_Settings_file (Path): export settings
        """
        simulation_settiings = toml.load(simulation_settiings_file)
        if export_Settings_file:
            export_Settings = toml.load(export_Settings_file)
        else:
            export_Settings = toml.load(
                simulation_settiings_file.parent / EXPORTS_SETTINGS_FILENAME
            )

        simulation_settiings = SimulationSettings(**simulation_settiings)
        export_Settings = ExportFileOptions(**export_Settings)
        return cls(simulation_settiings, export_Settings)

    def dump_settings(
        self,
        dest_dir: Path,
        simulation_file: str = SIMULATION_SETTINGS_FILENAME,
        export_file: str = EXPORTS_SETTINGS_FILENAME,
    ):
        """Dumps simulation settings to a provided path

        Args:
            dest_dir (Path): Directory where settins are dumped
            simulation_file (str, optional): simulation filename. Defaults to SIMULATION_SETTINGS_FILENAME.
            export_file (str, optional): export setting filename. Defaults to EXPORTS_SETTINGS_FILENAME.
        """

        settings_json = self.settings.model_dump_json()
        json.dump(settings_json, open(dest_dir / simulation_file, "w"))

        export_settings_json = self.settings.model_dump_json()
        json.dump(export_settings_json, open(dest_dir / export_file, "w"))

    def start_simulation(self):
        """Starts a loaded simulation

        Raises:
            Exception: Please pass a RAW or SAV file in the settings dictionary
        """

        self.hi = None
        self.simStartTime = time.time()

        if self.settings.simulation.case_study.exists():
            self.psse.case(str(self.settings.simulation.case_study))
        elif self.settings.simulation.raw_file.exists():
            self.psse.read(0, str(self.settings.simulation.raw_file))
        else:
            msg = "Please pass a RAW or SAV file in the settings dictionary"
            raise Exception(msg)

        logger.info(
            f"Trying to read a file >>{self.settings.simulation.case_study}"
        )
        self.raw_data = rd.Reader(self.psse)
        (
            self.bus_subsystems,
            self.all_subsysten_buses,
        ) = self.define_bus_subsystems()

        if self.export_settings.defined_subsystems_only:
            valid_buses = self.all_subsysten_buses
        else:
            valid_buses = self.raw_data.buses

        self.sim = sc.sim_controller(
            self.psse,
            self.dyntools,
            self.settings,
            self.export_settings,
            valid_buses,
            self.raw_data,
        )

        self.contingencies = self.build_contingencies()

        if self.settings.helics and self.settings.helics.cosimulation_mode:
            if self.settings.simulation.simulation_mode in [
                SimulationModes.DYNAMIC,
                SimulationModes.SNAP,
            ]:
                ...
            self.hi = HelicsInterface(
                self.psse,
                self.sim,
                self.settings,
                self.export_settings,
                self.bus_subsystems,
            )
            self.publications = self.hi.register_publications(
                self.bus_subsystems
            )
            if self.settings.helics.create_subscriptions:
                self.subscriptions = self.hi.register_subscriptions()

        if self.settings.simulation.gic_file:
            self.network_graph = self.parse_gic_file()
            self.bus_ids = self.network_graph.nodes.keys()
        else:
            self.network_graph = None

        self.results = Container(self.settings, self.export_settings)
        self.exp_vars = self.results.get_export_variables()
        self.inc_time = True

    def init(self):
        """Initializes the model"""

        self.sim.init(self.bus_subsystems)

        if self.settings.simulation.use_profile_manager:
            self.pm = ProfileManager(self.sim, self.settings)
            self.pm.setup_profiles()
        if self.settings.helics and self.settings.helics.cosimulation_mode:
            self.hi.enter_execution_mode()

    def parse_gic_file(self) -> Graph:
        """Parses the GIC file (if included in the project)

        Returns:
            Graph: Networkx graph representation for the model
        """

        gicdata = gp.GICParser(self.settings)
        return gicdata.psse_graph

    def define_bus_subsystems(self) -> (dict, list):
        """Defines a bussystem in the loaded PSSE model

        Raises:
            LookupError: Failed to create bus subsystem chosen buses.
            ValueError: Number of subsystems can not be more that 12. See PSSE documentation
            RuntimeError: Failed to add buses to bus subsystem

        Returns:
            dict: mapping of bus subsystems to buses
            list: List of bus subsystems
        """

        bus_subsystems_dict = {}
        bus_subsystems = self.get_bus_indices()
        # valid bus subsystem ID. Valid bus subsystem IDs range from 0 to 11 (PSSE documentation)
        if len(bus_subsystems) > MAX_PSSE_BUSSYSTEMS:
            msg = "Number of subsystems can not be more that 12. See PSSE documentation"
            raise ValueError(msg)

        all_subsysten_buses = []
        for i, buses in enumerate(bus_subsystems):
            if not buses:
                continue

            all_subsysten_buses.extend(buses)
            ierr = self.psse.bsysinit(i)
            if ierr:
                msg = "Failed to create bus subsystem chosen buses."
                raise LookupError(msg)
            else:
                logger.debug(f'Bus subsystem "{i}" created')

            ierr = self.psse.bsys(sid=i, numbus=len(buses), buses=buses)
            if ierr:
                msg = "Failed to add buses to bus subsystem."
                raise RuntimeError(msg)
            else:
                bus_subsystems_dict[i] = buses
                logger.debug(f'Buses {buses} added to subsystem "{i}"')
        all_subsysten_buses = [str(x) for x in all_subsysten_buses]
        return bus_subsystems_dict, all_subsysten_buses

    def get_bus_indices(self) -> BusSubsystems:
        """Returns bus indices for bus subsystems

        Returns:
            BusSubsystems: Bus subsystem model
        """

        if self.settings.bus_subsystems.from_file:
            bus_file = self.settings.bus_subsystems.bus_file
            bus_info = pd.read_csv(bus_file, index_col=None)
            bus_info = bus_info.values
            _, n_cols = bus_info.shape
            bus_data = []
            for col in range(n_cols):
                data = [int(x) for x in bus_info[:, col] if not np.isnan(x)]
                bus_data.append(data)
        else:
            bus_data = self.settings.bus_subsystems.bus_subsystem_list
        return bus_data

    def run(self):
        """Launches the simulation"""

        self._status = SimulationStatus.RUNNING_SIMULATION
        if self.sim.initialization_complete:
            if self.settings.plots and self.settings.plots.enable_dynamic_plots:
                bokeh_server_proc = subprocess.Popen(
                    ["bokeh", "serve"], stdout=subprocess.PIPE
                )  # noqa: S603,S607
            else:
                bokeh_server_proc = None

            logger.debug(
                f"Running dynamic simulation for time {self.settings.simulation.simulation_time.total_seconds()} sec"
            )
            total_simulation_time = (
                self.settings.simulation.simulation_time.total_seconds()
            )
            t = 0
            while True:
                self.step(t)
                if self.inc_time:
                    t += (
                        self.settings.simulation.simulation_step_resolution.total_seconds()
                    )
                if t >= total_simulation_time:
                    break

            self.psse.pssehalt_2()
            if not self.export_settings.export_results_using_channels:
                self.results.export_results()
            else:
                self.sim.export()

            if bokeh_server_proc is not None:
                bokeh_server_proc.terminate()
        else:
            logger.error(
                "Run init() command to initialize models before running the simulation"
            )
        self._status = "Simulation complete"

    def get_bus_ids(self) -> list:
        """Returns bus IDs

        Returns:
            list: Array of bus numbers
        """

        ierr, iarray = self.psse.abusint(-1, 1, "NUMBER")
        assert ierr == 0, f"Error code: {ierr}"
        return iarray

    def step(self, t: float) -> dict:
        """Steps through a single simulation time step. Is called iteratively to increment the simualtion

        Args:
            t (float): time step for the simulation

        Returns:
            dict: results from the current timestep
        """

        self.update_contingencies(t)
        if self.settings.simulation.use_profile_manager:
            self.pm.update()
        ctime = time.time() - self.simStartTime
        logger.debug(
            f"Simulation time: {t} seconds\nRun time: {ctime}\npsse time: {self.sim.get_time()}"
        )
        if self.settings.helics and self.settings.helics.cosimulation_mode:
            if self.settings.helics.create_subscriptions:
                self.update_subscriptions()
                logger.debug(f"Time requested: {t}")
                self.inc_time, helics_time = self.update_federate_time(t)
                logger.debug(f"Time granted: {helics_time}")

        if self.inc_time:
            self.sim.step(t)
        else:
            self.sim.resolve_step()

        if self.settings.helics and self.settings.helics.cosimulation_mode:
            self.publish_data()

        curr_results = self.update_result_container(t)
        return curr_results

    def update_result_container(self, t: float) -> dict:
        """Updates the result container with results from the current time step

        Args:
            t (float): simulation time in seconds

        Returns:
            dict: simulation reults from the current time step
        """

        if self.export_settings.defined_subsystems_only:
            curr_results = self.sim.read_subsystems(
                self.exp_vars, self.all_subsysten_buses
            )
        else:
            curr_results = self.sim.read_subsystems(
                self.exp_vars, self.raw_data.buses
            )

        if not USING_NAERM:
            if not self.export_settings.export_results_using_channels:
                self.results.update(
                    curr_results,
                    t,
                    self.sim.get_time(),
                    self.sim.has_converged(),
                )
        return curr_results

    def update_subscriptions(self):
        """Updates subscriptions (co-simulation mode only)"""

        self.hi.subscribe()

    def update_federate_time(self, t: float) -> (float, float):
        """Makes a time request to teh HELICS broker (co-simulation mode only)

        Args:
            t (float): simulation time in seconds

        Returns:
            float: requested time in seconds
            float: current simualtion time in seconds
        """

        inc_time, curr_time = self.hi.request_time(t)
        return inc_time, curr_time

    def publish_data(self):
        """Updates publications (co-simulation mode only)"""

        self.hi.publish()

    def get_results(self, params: Union[ExportAssetTypes, dict]) -> dict:
        """Returns queried simulation results

        Args:
            params (Union[ExportAssetTypes, dict]): _description_

        Returns:
            dict: simulation results
        """

        self._status = SimulationStatus.STARTING_RESULT_EXPORT
        self.exp_vars = self.results.update_export_variables(params)
        curr_results = (
            self.sim.read_subsystems(self.exp_vars, self.all_subsysten_buses)
            if self.export_settings.defined_subsystems_only
            else self.sim.read_subsystems(self.exp_vars, self.raw_data.buses)
        )
        self._status = SimulationStatus.RESULT_EXPORT_COMPLETE
        return curr_results

    def status(self) -> SimulationStatus:
        """returns current simulation status

        Returns:
            SimulationStatus: state of the simulator
        """
        return self._status.value

    def build_contingencies(
        self,
    ) -> List[
        Union[
            BusTripObject,
            BusFaultObject,
            LineTripObject,
            LineFaultObject,
            MachineTripObject,
        ]
    ]:
        """Builds user defined contengingies

        Returns:
            List[Union[BusFault, LineFault, LineTrip, BusTrip, MachineTrip]]: List of contingencies
        """

        contingencies = c.build_contingencies(self.psse, self.settings)
        return contingencies

    def inject_contingencies_external(self, contigencies: Contingencies):
        """Inject external contingencies.

        Args:
            contigencies (Contingencies): Contigencies Object
        """
        contingencies = c.build_contingencies(self.psse, contigencies)
        self.contingencies.extend(contingencies)

    def update_contingencies(self, t: float):
        """Updates contingencies during the simualtion run

        Args:
            t (float): simulation time in seconds
        """

        for contingency in self.contingencies:
            contingency.update(t)

    def force_psse_halt(self):
        """forces cleaup of pypss imort"""
        ierr = self.psse.pssehalt_2()
        assert ierr == 0, f"failed to halt PSSE. Error code - {ierr}"

    def __del__(self):
        if hasattr(self, "psse"):
            ierr = self.psse.pssehalt_2()
            assert ierr == 0, f"failed to halt PSSE. Error code - {ierr}"



# Builtin libraries
import logging
import zipfile
from contextlib import asynccontextmanager
from http import HTTPStatus

import uvicorn
from fastapi import FastAPI, UploadFile
from fastapi.exceptions import HTTPException
from fastapi.responses import HTMLResponse

from pypsse.api.common import BASE_PROJECT_PATH

# Internal libraries
from pypsse.api.web.handler import Handler
from pypsse.models import ApiPsseReply

logger = logging.getLogger(__name__)


class Server:
    def __init__(self):
        """Creates FAST API app to be served"""
        logger.info("Start server")
        self.handler = Handler()
        logger.info("Building web application")
        self.app = FastAPI(lifespan=self.lifespan)
        self.app.include_router(self.handler.router, prefix="/simulator")
        self.app.add_api_route("/", self.get_main_page, methods=["GET"])
        self.app.add_api_route("/upload", self.post_upload_zipped_project, methods=["POST"])
        self.app.add_api_route("/projects", self.get_list_projects, methods=["GET"])

        logger.info("Building API endpoints")

    async def get_main_page(self) -> HTMLResponse:
        """returns an html main page

        Returns:
            HTMLResponse: html page for base server url
        """

        html = """
        <h1>PyPSSE Server</h1>

        <h3>
        PyPSSE is a Python wrapper around psspy—a Python application programming
        interface (API) for the Power System Simulator for Engineering (PSS/E)—to perform
        time series power flow and dynamic simulation for power systems.
        </h3>

        <h3>
        The PSS/E Python API psspy follows functional programming methodology. The API
        exposes thousands of methods and can be difficult for new users to work with.
        PyPSSE wraps around hundreds of function calls in a few methods. This functionality
        allows users to set up cosimulations with minimal effort.
        </h3>

        <h3>For more information see check out the <a href="https://nrel.github.io/PyPSSE/">documentation</a></h3>

        <h3>For swagger documentation see: <a href="http://127.0.0.1:9090/docs/">SwaggerDocs</a></h3>

        <h3>For redoc documentation see: <a href="http://127.0.0.1:9090/redoc/">ReDoc</a></h3>

        """
        return HTMLResponse(html)

    @asynccontextmanager
    async def lifespan(self, *_, **__):
        """lifespan manager for the app. handeles logic for application startup and shutdown."""

        yield
        logger.info("cleanup_background_tasks")
        self.handler.shutdown_event.set()

    async def get_list_projects(self) -> list:
        """methods finds and returns pypsse projects

        Returns:
            list: returns a list of projects available on the server
        """

        folders = []
        for x in BASE_PROJECT_PATH.iterdir():
            if x.is_dir():
                folders.append(str(x.name))
        return {"folders": folders}

    async def post_upload_zipped_project(self, file: UploadFile):
        """upload a new zipped project to the server

        Args:
            file (UploadFile): _description_

        Raises:
            HTTPException: raised if invalid file exception
            HTTPException: raised if unexpected error happens server side

        Returns:
            ApiPsseReply: request returned
        """

        try:
            data = file.file.read()
            if not file.filename.endswith(".zip"):
                raise HTTPException(
                    status_code=HTTPStatus.NOT_ACCEPTABLE, detail="Invalid file type. Only zip files are accepted."
                )
            zip_filepath = BASE_PROJECT_PATH / file.filename
            project_path = BASE_PROJECT_PATH / file.filename.replace(".zip", "")
            project_path.mkdir(parents=True, exist_ok=True)

            with open(zip_filepath, "wb") as f:
                f.write(data)

            with zipfile.ZipFile(zip_filepath, "r") as zip_ref:
                zip_ref.extractall(project_path)

            zip_filepath.unlink()

            # TODO: update project name and psse path in settings

        except Exception as e:
            raise HTTPException(status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=str(e))
        return ApiPsseReply(status=HTTPStatus.OK, message=f"project upload to {project_path}")


def run_server(host: str, port: int):
    """Start the pypsse server

    Args:
        host (str): server ip adadress (set to switch between internal or external ip)
        port (int): server port
    """
    time_format = "%(asctime)s -  %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=time_format)
    # endpoints_file='app/endpoints.yaml'
    instance = Server()
    uvicorn.run(instance.app, host=host, port=port)



import numpy as np
from loguru import logger

from pypsse.models import ExportSettings, SimulationSettings
from pypsse.modes.abstract_mode import AbstractMode
from pypsse.modes.constants import DYNAMIC_ONLY_PPTY, converter, dyn_only_options
from pypsse.utils.dynamic_utils import DynamicUtils


class Dynamic(AbstractMode, DynamicUtils):
    "Class defination for dynamic simulation mode (uses dyr and raw files)"

    def __init__(
        self,
        psse,
        dyntools,
        settings: SimulationSettings,
        export_settings: ExportSettings,
        subsystem_buses,
        raw_data,
    ):
        super().__init__(psse, dyntools, settings, export_settings, subsystem_buses, raw_data)
        self.time = settings.simulation.start_time
        self._StartTime = settings.simulation.start_time
        self.incTime = settings.simulation.simulation_step_resolution

        self.init({})

    def init(self, bus_subsystems):
        "Initializes the simulation"
        super().init(bus_subsystems)
        self.iter_const = 100.0

        if self.settings.simulation.rwm_file:
            self.psse.mcre([1, 0], self.rwn_file)

        self.psse.fnsl([0, 0, 0, 1, 0, 0, 0, self._i])

        self.load_setup_files()
        self.convert_load()

        self.psse.gnet(1, 0)
        self.psse.fdns([1, 1, 0, 1, 1, 0, 0, 0])
        self.psse.fnsl([1, 1, 0, 1, 1, 0, 0, 0])
        self.psse.cong(0)
        # Solve for dynamics
        self.psse.ordr(0)
        self.psse.fact()
        self.psse.tysl(0)
        self.psse.tysl(0)
        # self.psse.save(self.study_case_path.split('.')[0] + ".sav")
        dyr_path = self.settings.simulation.dyr_file
        assert dyr_path and dyr_path.exists
        logger.debug(f"Loading dynamic model....{dyr_path}")
        self.psse.dynamicsmode(1)
        ierr = self.psse.dyre_new([1, 1, 1, 1], str(dyr_path), r"""conec""", r"""conet""", r"""compile""")

        if self.settings.helics and self.settings.helics.cosimulation_mode:
            if self.settings.helics.iterative_mode:
                sim_step = self.settings.simulation.psse_solver_timestep.total_seconds() / self.iter_const
            else:
                sim_step = self.settings.simulation.psse_solver_timestep.total_seconds()
        else:
            sim_step = self.settings.simulation.psse_solver_timestep.total_seconds()

        ierr = self.psse.dynamics_solution_param_2(
            [60, self._i, self._i, self._i, self._i, self._i, self._i, self._i],
            [0.4, self._f, sim_step, self._f, self._f, self._f, self._f, self._f],
        )

        if ierr:
            msg = f'Error loading dynamic model file "{dyr_path}". Error code - {ierr}'
            raise Exception(msg)
        else:
            logger.debug(f"Dynamic file {dyr_path} sucessfully loaded")

        self.disable_load_models_for_coupled_buses()

        if self.export_settings.export_results_using_channels:
            self.setup_channels()

        self.psse.delete_all_plot_channels()

        self.setup_all_channels()

        # Load user defined models
        self.load_user_defined_models()

        # Load flow settings
        self.psse.fdns([0, 0, 0, 1, 1, 0, 99, 0])
        # initialize
        ierr = self.psse.strt_2(
            [
                1,
                self.settings.generators.missing_machine_model,
            ],
            str(self.settings.export.outx_file),
        )
        if ierr:
            self.initialization_complete = False
            msg = f"Dynamic simulation failed to successfully initialize. Error code - {ierr}"
            raise Exception(msg)
        else:
            self.initialization_complete = True
            logger.debug("Dynamic simulation initialization sucess!")
        # get load info for the sub system
        self.load_info = self.get_load_indices(bus_subsystems)

        logger.debug("pyPSSE initialization complete!")

        self.xTime = 0

        return self.initialization_complete

    def step(self, t):
        "Increments the simulation"

        self.time = self.time + self.incTime
        self.xTime = 0
        return self.psse.run(0, t, 1, 1, 1)

    def get_load_indices(self, bus_subsystems):
        "Returns load indices"

        all_bus_ids = {}
        for bus_subsystem_id in bus_subsystems.keys():
            load_info = {}
            ierr, load_data = self.psse.aloadchar(bus_subsystem_id, 1, ["ID", "NAME", "EXNAME"])

            load_data = np.array(load_data)
            ierr, bus_data = self.psse.aloadint(bus_subsystem_id, 1, ["NUMBER"])

            bus_data = bus_data[0]
            for i, bus_id in enumerate(bus_data):
                load_info[bus_id] = {
                    "Load ID": load_data[0, i],
                    "Bus name": load_data[1, i],
                    "Bus name (ext)": load_data[2, i],
                }
            all_bus_ids[bus_subsystem_id] = load_info
        return all_bus_ids

    def resolve_step(self, t):
        "Resolves the current time step"

        err = self.psse.run(0, t + self.xTime * self.incTime / self.iter_const, 1, 1, 1)
        self.xTime += 1
        return err

    def get_time(self):
        "Returns current simulator time"

        return self.time

    def get_total_seconds(self):
        "Returns total simulation time"

        return (self.time - self._StartTime).total_seconds()

    def get_step_size_cec(self):
        "Returns simulation timestep resolution"
        return self.settings.simulation.simulation_step_resolution.total_seconds()

    @converter
    def read_subsystems(self, quantities, subsystem_buses, ext_string2_info=None, mapping_dict=None):
        "Queries the result container for current results"

        if ext_string2_info is None:
            ext_string2_info = {}
        if mapping_dict is None:
            mapping_dict = {}
        results = super().read_subsystems(
            quantities, subsystem_buses, mapping_dict=mapping_dict, ext_string2_info=ext_string2_info
        )

        poll_results = self.poll_channels()
        results.update(poll_results)
        for class_name, var_list in quantities.items():
            if class_name in dyn_only_options:
                for v in var_list:
                    if v in DYNAMIC_ONLY_PPTY[class_name]:
                        for func_name in dyn_only_options[class_name]:
                            if v in dyn_only_options[class_name][func_name]:
                                con_ind = dyn_only_options[class_name][func_name][v]
                                for bus in subsystem_buses:
                                    if class_name == "Loads":
                                        ierr = self.psse.inilod(int(bus))

                                        ierr, ld_id = self.psse.nxtlod(int(bus))

                                        if ld_id is not None:
                                            ierr, con_index = getattr(self.psse, func_name)(
                                                int(bus), ld_id, "CHARAC", "CON"
                                            )

                                            if con_index is not None:
                                                act_con_index = con_index + con_ind
                                                ierr, value = self.psse.dsrval("CON", act_con_index)

                                                res_base = f"{class_name}_{v}"
                                                if res_base not in results:
                                                    results[res_base] = {}
                                                obj_name = f"{bus}_{ld_id}"
                                                results[res_base][obj_name] = value
            else:
                logger.warning("Extend function 'read_subsystems' in the Dynamic class (Dynamic.py)")
        return results



"""Defines all standard shared varibales in PyPSSE"""

VALUE_UPDATE_BOUND = 1e6
MAX_PSSE_BUSSYSTEMS = 12

CASESTUDY_FOLDER = "case_study"
SETTINGS_FOLDER = "settings"
EXPORTS_FOLDER = "exports"
LOGS_FOLDER = "logs"
PROFILES_FOLDER = "profiles"
DEFAULTS_FOLDER = "defaults"
EXAMPLES_FOLDER = "examples"

SIMULATION_SETTINGS_FILENAME = "simulation_settings.toml"
EXPORTS_SETTINGS_FILENAME = "export_settings.toml"
DEFAULT_SUBSCRIPTION_FILENAME = "subscriptions.csv"
DEFAULT_PROFILE_MAPPING_FILENAME = "profile_mapping.toml"
DEFAULT_PROFILE_STORE_FILENAME = "profiles.hdf5"
DEFAULT_RESULTS_FILENAME = "simulation_results.hdf5"


DEFAULT_OUT_FILE = "results.out"
DEFAULT_OUTX_FILE = "results.outx"
DEFAULT_EXCEL_FILE = "results.xls"
DEFAULT_LOG_FILE = "psse.log"
DEFAULT_GRAPH_FILE = "network.gpickle"
DEFAULT_COORDINATES_FILE = "coordinates.csv"

MAPPED_CLASS_NAMES = {
    "buses": "Buses",
    "areas": "Areas",
    "zones": "Zones",
    "loads": "Loads",
    "branches": "Branches",
    "machines": "Machines",
    "stations": "Stations",
    "transformers": "Transformers",
    "fixed_shunts": "Fixed_shunts",
    "switched_shunts": "Switched_shunts",
    "induction_generators": "Induction_generators",
    "dctransmissionlines": "DCtransmissionlines",
}

MACHINE_CHANNELS = {
    "ANGLE": 1,
    "PELEC": 2,
    "QELEC": 3,
    "ETERM": 4,
    "EFD": 5,
    "PMECH": 6,
    "SPEED": 7,
    "XADIFD": 8,
    "ECOMP": 9,
    "VOTHSG": 10,
    "VREF": 11,
    "VUEL": 12,
    "VOEL": 13,
    "GREF": 14,
    "LCREF": 15,
    "WVLCTY": 16,
    "WTRBSP": 17,
    "WPITCH": 18,
    "WAEROT": 19,
    "WROTRV": 20,
    "WROTRI": 21,
    "WPCMND": 22,
    "WQCMND": 23,
    "WAUXSG": 24,
}



"""
CLI to run a PyDSS project
"""

from pathlib import Path

import click
import toml

from pypsse.common import EXPORTS_SETTINGS_FILENAME, SIMULATION_SETTINGS_FILENAME
from pypsse.models import ExportFileOptions, SimulationSettings
from pypsse.simulator import Simulator


@click.argument(
    "project-path",
)
@click.option(
    "-s",
    "--simulations-file",
    required=False,
    default=SIMULATION_SETTINGS_FILENAME,
    show_default=True,
    help="scenario toml file to run (over rides default)",
)
@click.command()
def run(project_path, simulations_file=None):
    """Runs a valid PyPSSE simulation."""
    file_path = Path(project_path) / simulations_file
    msg = "Simulation file not found. Use -s to choose a valid settings file"
    "if its name differs from the default file name."
    assert file_path.exists(), msg
    x = Simulator.from_setting_files(file_path)
    x.run()



import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from http import HTTPStatus
from multiprocessing import Event, Process, Queue, cpu_count
from typing import Any
from uuid import uuid4

from fastapi import APIRouter, WebSocket
from fastapi.exceptions import HTTPException
from fastapi.responses import FileResponse
from pydantic import UUID4

from pypsse.api.app.psse import SimulatorAPI, SimulatorWebSocket
from pypsse.api.common import BASE_PROJECT_PATH
from pypsse.common import DEFAULT_LOG_FILE, DEFAULT_RESULTS_FILENAME, EXPORTS_FOLDER, LOGS_FOLDER
from pypsse.models import ApiPsseException, ApiPssePostRequest, ApiPssePutRequest, ApiPsseReply, ApiPsseReplyInstances

logger = logging.getLogger(__name__)

shutdown_event = Event()


class Handler:

    """Handlers for web server."""

    uuid_to_project_mapping = {}

    def __init__(self):
        """Constructor for PSSE handler."""

        logger.info("Initializing service handler")

        # Initializing dictionary for stroing psse_instances
        self.psse_instances = {}

        # Event flag to control shutdown of background tasks
        self.shutdown_event = Event()

        # Pool and Loop are used to catch returning Multiprocessing Queue messages
        #  to prevent blocking of the API.

        self.pool = ThreadPoolExecutor(max_workers=cpu_count() - 1)
        self.loop = asyncio.get_event_loop()
        self.router = APIRouter()
        self.router.add_websocket_route("/ws", self.handle_psse_websocket)
        self.router.add_api_route("/", self.post_psse, methods=["POST"])
        self.router.add_api_route("/", self.put_psse, methods=["PUT"])
        self.router.add_api_route("/uuid/{uuid}", self.delete_psse, methods=["DELETE"])
        self.router.add_api_route("/uuids", self.get_instance_uuids, methods=["GET"])
        self.router.add_api_route("/status/uuid/{uuid}", self.get_instance_status, methods=["GET"])
        self.router.add_api_route("/results/uuid/{uuid}", self.get_download_results, methods=["GET"])
        self.router.add_api_route("/log/uuid/{uuid}", self.get_download_logs, methods=["GET"])

    async def handle_psse_websocket(self, websocket: WebSocket):
        """Function to handle PSSE websocket."

        Args:
            websocket (WebSocket): web socket object
        """

        to_psse_queue, from_psse_queue = Queue(), Queue()
        psse_uuid = str(uuid4())

        p = Process(
            target=SimulatorWebSocket,
            name=psse_uuid,
            args=(
                shutdown_event,
                to_psse_queue,
                from_psse_queue,
            ),
        )
        p.start()

        logger.info("Starting websocket communication for %s", psse_uuid)

        await websocket.accept()
        result = from_psse_queue.get()
        await websocket.send_json(result)

        while True:
            data = await websocket.receive_text()
            logger.info("websocket message data: %s", str(result)[1:1000])
            to_psse_queue.put(data)
            result = from_psse_queue.get()
            logger.info("psse handler output = %s", str(result)[1:1000])
            await websocket.send_json(result)

            if data == "END":
                await websocket.close()
                break

    async def post_psse(self, request: ApiPssePostRequest) -> ApiPsseReply:
        """Create UUID and intialize and push to queue

        Args:
            request (ApiPssePostRequest): request for the server

        Returns:
            ApiPsseReply: returns response from the server
        """

        psse_uuid = str(uuid4())
        self.uuid_to_project_mapping[psse_uuid] = request.project_name
        to_psse_queue = Queue()
        from_psse_queue = Queue()

        # Create a process for launching PSSE instance
        p = Process(
            target=SimulatorAPI, name=psse_uuid, args=(self.shutdown_event, to_psse_queue, from_psse_queue, request)
        )

        # Store queue and process
        self.psse_instances[psse_uuid] = {
            "to_psse_queue": to_psse_queue,
            "from_psse_queue": from_psse_queue,
            "process": p,
        }

        # Catching data coming from PSSE
        psse_t = self.loop.run_in_executor(self.pool, self._post_put_background_task, psse_uuid)
        psse_t.add_done_callback(self._post_put_callback)

        # Start process for psse
        try:
            p.start()

            # Return a message to webclient
            return ApiPsseReply(
                status=HTTPStatus.OK,
                message="Starting PSSE",
                uuid=psse_uuid,
            )
        except Exception as e:
            self._base_error(HTTPStatus.INTERNAL_SERVER_ERROR, str(e), psse_uuid)

    async def put_psse(self, request: ApiPssePutRequest) -> ApiPsseReply:
        """Method for simulator put requst

        Args:
            request (ApiPssePutRequest): request object for the server

        Returns:
            ApiPsseReply: response from the server
        """
        logger.info(f"Running command :{request.model_dump_json()}")
        self._error_invalid_uuid(request.uuid)
        if not hasattr(request, "command") or not hasattr(request, "parameters"):
            self._base_error(HTTPStatus.NOT_FOUND, "Please provide a command and parameters", request.uuid)

        # psse_t = self.loop.run_in_executor(self.pool, self._post_put_background_task, request.uuid)
        # psse_t.add_done_callback(self._post_put_callback)

        logger.info(f"Submitted data to psse :{request.model_dump_json()}")
        self.psse_instances[str(request.uuid)]["to_psse_queue"].put(request.model_dump_json())
        results = self.psse_instances[str(request.uuid)]["from_psse_queue"].get()

        return ApiPsseReply(status="", message=f"{results}", uuid=request.uuid)

    async def delete_psse(self, uuid: UUID4) -> ApiPsseReply:
        """Delete an instance of simulation

        Args:
            uuid (UUID4): simulation instance uuid

        Returns:
            ApiPsseReply: response from the server
        """
        """Delete an instance of simulation"""
        self._error_invalid_uuid(uuid)
        try:
            psse_t = self.loop.run_in_executor(self.pool, self._delete_background_task, uuid)
            psse_t.add_done_callback(self._delete_callback)
            self.psse_instances[str(uuid)]["to_psse_queue"].put("END")
            self.psse_instances.pop(str(uuid))
            self.uuid_to_project_mapping.pop(str(uuid))
            return ApiPsseReply(status="", message="Attempting to close PyPSSE instance", uuid=uuid)
        except Exception:
            msg = f"Error closing in PSSE instance {uuid}"
            logger.error(msg)
            self._base_error(HTTPStatus.INTERNAL_SERVER_ERROR, "Error closing in PyPSSE instance", uuid)

    async def get_instance_uuids(self) -> ApiPsseReplyInstances:
        """get all running simulation uuids

        Returns:
            ApiPsseReplyInstances: server response
        """

        uuids = [str(k) for k in self.psse_instances.keys()]
        return ApiPsseReplyInstances(status=HTTPStatus.OK, message=f"{len(uuids)} uuids found", simulators=uuids)

    async def get_instance_status(self, uuid: UUID4) -> ApiPsseReplyInstances:
        """get status of the current provided simuation instance

        Args:
            uuid (UUID4): simulation instance UUID

        Returns:
            ApiPsseReplyInstances: server response
        """

        self._error_invalid_uuid(uuid)
        return ApiPsseReplyInstances(
            status=HTTPStatus.OK,
            message="UUID exists",
            uuid=str(uuid),
        )

    async def get_download_results(self, uuid: UUID4) -> FileResponse:
        """download results from a simulation instance

        Args:
            uuid (UUID4): simulation instance UUID

        Returns:
            FileResponse: hdf5 simulation results
        """

        self._error_invalid_uuid(uuid)

        project_name = self.uuid_to_project_mapping[str(uuid)]
        results_file = BASE_PROJECT_PATH / project_name / EXPORTS_FOLDER / DEFAULT_RESULTS_FILENAME
        if not results_file.exists():
            self._base_error(HTTPStatus.NOT_FOUND, "File not found", uuid)

        return FileResponse(path=results_file, filename=DEFAULT_RESULTS_FILENAME, media_type="hdf5")

    async def get_download_logs(self, uuid: UUID4) -> FileResponse:
        """download logs from a simulation instance

        Args:
            uuid (UUID4): simlation instance UUID

        Returns:
            FileResponse: returns simulation log files
        """

        if str(uuid) not in self.psse_instances.keys():
            param = "UUID={psse_uuid} not found in the PSSE instances"
            logger.error(f"{param}")
            self._base_error(HTTPStatus.NOT_FOUND, param, uuid)

        project_name = self.uuid_to_project_mapping[str(uuid)]
        results_file = BASE_PROJECT_PATH / project_name / LOGS_FOLDER / DEFAULT_LOG_FILE
        if not results_file.exists():
            self._base_error(HTTPStatus.NOT_FOUND, f"File not found - {results_file}", uuid)

        return FileResponse(path=results_file, filename=DEFAULT_LOG_FILE, media_type="hdf5")

    def _post_put_background_task(self, psse_uuid: UUID4) -> Any:
        """background task to retrieve results from simulation queue

        Args:
            psse_uuid (UUID4): simulation instance UUID

        Returns:
            Any: data types returned from the server
        """
        q = self.psse_instances[psse_uuid]["from_psse_queue"]
        ans = q.get()
        return ans

    def _post_put_callback(self, return_value):
        logger.info(f"{return_value.result()}")

    def _get_uuid(self, data: Any) -> UUID4:
        """returns simulation UUID

        Args:
            data (Any): pypsse objects

        Returns:
            UUID4: simulation UUID returned
        """

        if not hasattr(data, "uuid"):
            return None

        psse_uuid = data.uuid
        if psse_uuid not in self.psse_instances.keys():
            return None

        return psse_uuid

    def _delete_background_task(self, psse_uuid: UUID4) -> dict:
        """_summary_

        Args:
            psse_uuid (UUID4): simuation instance UUID

        Returns:
            dict: method response
        """
        while self.psse_instances[psse_uuid]["process"].is_alive():
            continue

        del self.psse_instances[psse_uuid]

        return {"Status": "Success", "Message": "PSSE instance closed", "UUID": psse_uuid}

    def _delete_callback(self, return_value):
        logger.info(f"{return_value.result()}")

    def _error_invalid_uuid(self, uuid: UUID4):
        if not uuid or str(uuid) not in self.psse_instances:
            self._base_error(HTTPStatus.NOT_FOUND, "UUID does not exist", uuid)

    def _base_error(self, status_code: HTTPStatus, message: str, uuid: UUID4):
        """_summary_

        Args:
            status_code (HTTPStatus): simulation status
            message (str): message from simulator
            uuid (UUID4): simulation instance UUID

        Raises:
            HTTPException: _description_
        """
        raise HTTPException(
            status_code=status_code,
            detail=ApiPsseException(
                message=message,
                uuid=uuid,
            ).model_dump_json(),
        )



import datetime
from typing import Union

import pandas as pd
from loguru import logger

from pypsse.common import EXPORTS_FOLDER, MAPPED_CLASS_NAMES
from pypsse.data_writers.data_writer import DataWriter
from pypsse.enumerations import BulkWriteModes, StreamedWriteModes
from pypsse.models import ExportAssetTypes, ModelTypes, SimulationSettings


class Container:
    "Class defination for the simulation result container"

    BULK_WRITE_MODES = [m.value for m in BulkWriteModes]
    STREAMED_WRITE_MODES = [m.value for m in StreamedWriteModes]

    def __init__(self, settings: SimulationSettings, export_settings: ExportAssetTypes):
        """Sets up the result container object

        Args:
            settings (SimulationSettings): _description_
            export_settings (ExportAssetTypes): _description_
        """

        export__list = [m.value for m in ModelTypes]
        self.export_path = settings.simulation.project_path / EXPORTS_FOLDER
        self.export_settings = export_settings
        self.settings = settings
        self.results = {}
        self.export_vars = {}
        for class_name in export__list:
            mapped_name = MAPPED_CLASS_NAMES[class_name.lower()]
            variables = getattr(export_settings, class_name.lower())
            if variables:
                for variable in variables:
                    self.results[f"{mapped_name}_{variable.value}"] = None
                    if mapped_name not in self.export_vars:
                        self.export_vars[mapped_name] = []
                    self.export_vars[mapped_name].append(variable.value)

        time_steps = int(
            self.settings.simulation.simulation_time.total_seconds()
            / self.settings.simulation.simulation_step_resolution.total_seconds()
        )
        if self.export_settings.file_format not in self.BULK_WRITE_MODES:
            self.dataWriter = DataWriter(self.export_path, export_settings.file_format.value, time_steps)

    def update_export_variables(self, params: Union[ExportAssetTypes, dict]) -> dict:
        """Updates the container with current system state.
        Method is called iteratively to store results as a simulation executes

        Args:
            params (Union[ExportAssetTypes, dict]): _description_

        Returns:
            dict: mapping of export variables to values
        """
        export__list = [m.value for m in ModelTypes]
        self.results = {}
        self.export_vars = {}
        if params:
            if isinstance(params, ExportAssetTypes):
                class_assets = params
            else:
                class_assets = ExportAssetTypes(**params)
        else:
            class_assets = self.export_settings

        for class_name in export__list:
            if class_name in params:
                mapped_name = MAPPED_CLASS_NAMES[class_name]
                variables = getattr(class_assets, class_name)
                if variables:
                    for variable in variables:
                        self.results[f"{mapped_name}_{variable.value}"] = None
                        if mapped_name not in self.export_vars:
                            self.export_vars[mapped_name] = []
                        self.export_vars[mapped_name].append(variable.value)

        return self.export_vars

    def get_export_variables(self) -> dict:
        """Queries and return results from the current timestep

        Returns:
            dict: mapping of export variables to values
        """

        return self.export_vars

    def update(self, bus_data: dict, _, time: datetime.datetime, has_converged: bool):
        """Updates the results cotainer

        Args:
            bus_data (dict): mapping of vairables to values
            _ (_type_): _description_
            time (datetime.datetime): simulation time
            has_converged (bool): flag showing if simulation converged
        """

        if self.export_settings.file_format not in self.BULK_WRITE_MODES:
            self.dataWriter.write(time, bus_data, has_converged)
        else:
            for variable_name, _ in bus_data.items():
                if not isinstance(self.results[f"{variable_name}"], pd.DataFrame):
                    self.results[f"{variable_name}"] = pd.DataFrame(bus_data[variable_name], index=[0])
                else:
                    df1 = self.results[f"{variable_name}"]
                    df2 = pd.DataFrame.from_dict([bus_data[variable_name]])
                    concatenated = pd.concat([df1, df2])
                    self.results[f"{variable_name}"] = concatenated
        logger.debug("result container updated")

    def export_results(self):
        """exports all results stored to an external file"""

        if self.export_settings.file_format in self.BULK_WRITE_MODES:
            for df_name, df in self.results.items():
                export_path = (
                    self.settings.simulation.project_path
                    / EXPORTS_FOLDER
                    / f'{df_name}.{self.export_settings["Write format"]}'
                )
                if self.export_settings.file_format == BulkWriteModes.CSV:
                    if isinstance(df, pd.DataFrame):
                        df.to_csv(export_path)
                elif self.export_settings.file_format == BulkWriteModes.PKL:
                    df.to_pickle(export_path)
                logger.info(f"results exported to {export_path}")



import ast

import helics as h
import pandas as pd
from loguru import logger

from pypsse.common import MAPPED_CLASS_NAMES, VALUE_UPDATE_BOUND
from pypsse.models import ExportFileOptions, SimulationSettings
from pypsse.modes.abstract_mode import AbstractMode
from pypsse.profile_manager.common import PROFILE_VALIDATION


class HelicsInterface:
    "Implements the HEILCS interface for PyPSSE"
    all_sub_results = {}
    all_pub_results = {}

    dynamic_iter_const = 1000.0
    n_states = 5
    init_state = 1
    dynamic_params = ("FmA", "FmB", "FmC", "FmD", "Fel")

    def __init__(
        self,
        psse: object,
        sim: AbstractMode,
        settings: SimulationSettings,
        export_settings: ExportFileOptions,
        bus_subsystems: dict,
    ):
        """Sets up the co-simulation federate

        Args:
            psse (object): simulator instance
            sim (AbstractMode): simulator controller inatance
            settings (SimulationSettings): simulation settings
            export_settings (ExportFileOptions): export settings
            bus_subsystems (dict): bus subsystem to bus mapping
        """
        self.bus_pubs = ["bus_id", "bus_Vmag", "bus_Vang", "bus_dev"]
        self.psse = psse
        self.settings = settings
        self.export_settings = export_settings
        self.bus_subsystems = bus_subsystems
        self.sim = sim
        self.c_seconds = 0
        self.c_seconds_old = -1

        self._co_convergance_error_tolerance = settings.helics.error_tolerance
        self._co_convergance_max_iterations = settings.helics.max_coiterations
        self.create_federate()
        self.subsystem_info = []
        self.publications = {}
        self.subscriptions = {}

    def enter_execution_mode(self):
        """Enables federate to enter execution mode"""
        itr = 0
        itr_flag = h.helics_iteration_request_iterate_if_needed
        while True:
            itr_status = h.helicsFederateEnterExecutingModeIterative(self.psse_federate, itr_flag)
            logger.debug(f"--- Iter {itr}: Iteration Status = {itr_status}, Passed Iteration Requestion = {itr_flag}")
            if itr_status == h.helics_iteration_result_next_step:
                break

    def create_federate(self):
        """Creates a HELICS co-simulation federate"""
        self.fedinfo = h.helicsCreateFederateInfo()
        h.helicsFederateInfoSetCoreName(self.fedinfo, self.settings.helics.federate_name)
        h.helicsFederateInfoSetCoreTypeFromString(self.fedinfo, self.settings.helics.core_type.value)
        h.helicsFederateInfoSetCoreInitString(self.fedinfo, "--federates=1")
        h.helicsFederateInfoSetBroker(self.fedinfo, str(self.settings.helics.broker_ip))
        h.helicsFederateInfoSetBrokerPort(self.fedinfo, self.settings.helics.broker_port)

        if self.settings.helics.iterative_mode:
            h.helicsFederateInfoSetTimeProperty(
                self.fedinfo,
                h.helics_property_time_delta,
                self.settings.simulation.simulation_step_resolution.total_seconds() / self.dynamic_iter_const,
            )
        else:
            h.helicsFederateInfoSetTimeProperty(
                self.fedinfo,
                h.helics_property_time_delta,
                self.settings.simulation.simulation_step_resolution.total_seconds(),
            )
        self.psse_federate = h.helicsCreateValueFederate(self.settings.helics.federate_name, self.fedinfo)

    def register_publications(self, bus_subsystems: dict):
        """Creates a HELICS publications

        Args:
            bus_subsystems (dict): bus subsystem to bus mapping

        Raises:
            LookupError: returned if invalid subsystem passed
            LookupError: returned if invalid class named passed
            LookupError: returned if property fields do not match class names
        """
        self.publications = {}
        self.pub_struc = []
        for publication_dict in self.settings.helics.publications:
            bus_subsystem_ids = publication_dict.bus_subsystems
            if not set(bus_subsystem_ids).issubset(self.bus_subsystems):
                msg = f"One or more invalid bus subsystem ID pass in {bus_subsystem_ids}."
                f"Valid subsystem IDs are  '{list(self.bus_subsystems.keys())}'."
                raise LookupError(msg)

            elm_class = publication_dict.asset_type.value

            if not hasattr(self.export_settings, elm_class.lower()):
                msg = f"'{elm_class.lower()}' is not a valid class of elements. "
                f"Valid fields are: {list(self.export_settings.dict().keys())}"
                raise LookupError(msg)

            managed_properties = [ptpy.value for ptpy in getattr(self.export_settings, elm_class.lower())]
            properties = [p.value for p in publication_dict.asset_properties]

            if not set(properties).issubset(managed_properties):
                msg = f"One or more publication property defined for class '{elm_class}' is invalid."
                " Properties defined {properties}"
                f"Valid properties for class '{elm_class.lower()}' are '{managed_properties}'"

                raise Exception(msg)

            bus_cluster = []
            for bus_subsystem_id in bus_subsystem_ids:
                bus_cluster.extend([str(x) for x in bus_subsystems[bus_subsystem_id]])

            if elm_class.lower() in MAPPED_CLASS_NAMES:
                elm_class = MAPPED_CLASS_NAMES[elm_class.lower()]

            self.pub_struc.append([{elm_class: properties}, bus_cluster])
            temp_res = self.sim.read_subsystems({elm_class: properties}, bus_cluster)
            temp_res = self.get_restructured_results(temp_res)
            for c_name, elm_info in temp_res.items():
                for name, v_info in elm_info.items():
                    for p_name, val in v_info.items():
                        pub_tag = f"{self.settings.helics.federate_name}.{c_name}.{name}.{p_name}"
                        dtype_matched = True
                        if isinstance(val, float):
                            self.publications[pub_tag] = h.helicsFederateRegisterGlobalTypePublication(
                                self.psse_federate, pub_tag, "double", ""
                            )
                        elif isinstance(val, complex):
                            self.publications[pub_tag] = h.helicsFederateRegisterGlobalTypePublication(
                                self.psse_federate, pub_tag, "complex", ""
                            )
                        elif isinstance(val, int):
                            self.publications[pub_tag] = h.helicsFederateRegisterGlobalTypePublication(
                                self.psse_federate, pub_tag, "integer", ""
                            )
                        elif isinstance(val, list):
                            self.publications[pub_tag] = h.helicsFederateRegisterGlobalTypePublication(
                                self.psse_federate, pub_tag, "vector", ""
                            )
                        elif isinstance(val, str):
                            self.publications[pub_tag] = h.helicsFederateRegisterGlobalTypePublication(
                                self.psse_federate, pub_tag, "string", ""
                            )
                        else:
                            dtype_matched = False
                            logger.warning(f"Publication {pub_tag} could not be registered. Data type not found")
                        if dtype_matched:
                            logger.debug(f"Publication registered: {pub_tag}")

    def register_subscriptions(self):
        """Creates a HELICS subscriptions

        Raises:
            Exception: raised if invalid element type passed
            Exception: raised if property do not match class
            Exception: rasied if invalid property
        """

        self.subscriptions = {}
        assert (
            self.settings.simulation.subscriptions_file
        ), "HELICS co-simulations requires a subscriptions_file property populated"
        sub_data = pd.read_csv(self.settings.simulation.subscriptions_file)
        self.psse_dict = {}
        for _, row in sub_data.iterrows():
            try:
                row["element_property"] = ast.literal_eval(row["element_property"])
            except Exception as e:
                logger.debug(str(e))
            try:
                row["scaler"] = ast.literal_eval(row["scaler"])
            except Exception as e:
                logger.debug(str(e))

            if row["element_type"] not in PROFILE_VALIDATION:
                msg = f"Subscription file error: {row['element_type']} not a valid element_type."
                f"Valid element_type are: {list(PROFILE_VALIDATION.keys())}"
                raise Exception(msg)

            if (
                isinstance(row["element_property"], str)
                and row["element_property"] not in PROFILE_VALIDATION[row["element_type"]]
            ):
                msg = f"Subscription file error: {row['property']} is not valid. "
                f"Valid subtypes for '{row['element_type']}' are: {PROFILE_VALIDATION[row['element_type']]}"
                raise Exception(msg)

            valid_set = set(PROFILE_VALIDATION[row["element_type"]])
            s_set = set(row["element_property"])
            if isinstance(row["element_property"], list) and valid_set.issubset(s_set):
                msg = f"Subscription file error: {row['element_property']} is not a valid subset. "
                f"Valid subtypes for '{row['element_type']}' are: {PROFILE_VALIDATION[row['element_type']]}"
                raise Exception(msg)

            element_id = str(row["element_id"])

            self.subscriptions[row["sub_tag"]] = {
                "bus": row["bus"],
                "element_id": element_id,
                "element_type": row["element_type"],
                "property": row["element_property"],
                "scaler": row["scaler"],
                "dStates": [self.init_state] * self.n_states,
                "subscription": h.helicsFederateRegisterSubscription(self.psse_federate, row["sub_tag"], ""),
            }

            logger.info(
                "{} property of element {}.{} at bus {} has subscribed to {}".format(
                    row["element_property"], row["element_type"], row["element_id"], row["bus"], row["sub_tag"]
                )
            )

            if row["bus"] not in self.psse_dict:
                self.psse_dict[row["bus"]] = {}
            if row["element_type"] not in self.psse_dict[row["bus"]]:
                self.psse_dict[row["bus"]][row["element_type"]] = {}
            if element_id not in self.psse_dict[row["bus"]][row["element_type"]]:
                self.psse_dict[row["bus"]][row["element_type"]][element_id] = {}
            if isinstance(row["element_property"], str):
                if row["element_property"] not in self.psse_dict[row["bus"]][row["element_type"]][element_id]:
                    self.psse_dict[row["bus"]][row["element_type"]][element_id][row["element_property"]] = 0
            elif isinstance(row["element_property"], list):
                for r in row["element_property"]:
                    if r not in self.psse_dict[row["bus"]][row["element_type"]][element_id]:
                        self.psse_dict[row["bus"]][row["element_type"]][element_id][r] = 0

    def request_time(self, _) -> (bool, float):
        """Enables time increment of the federate ina  co-simulation."
        Works for both loosely ans tightly coupled co-simulations

        Args:
            self (_type_): _description_
            float (_type_): _description_

        Returns:
            bool: flag for iteration requrirement (rerun same time step)
            float: current helics time in seconds
        """

        r_seconds = self.sim.get_total_seconds()  # - self._dss_solver.GetStepResolutionSeconds()
        if self.sim.get_time() not in self.all_sub_results:
            self.all_sub_results[self.sim.get_time()] = {}
            self.all_pub_results[self.sim.get_time()] = {}

        if not self.settings.helics.iterative_mode:
            while self.c_seconds < r_seconds:
                self.c_seconds = h.helicsFederateRequestTime(self.psse_federate, r_seconds)
            logger.info(f"Time requested: {r_seconds} - time granted: {self.c_seconds} ")
            return True, self.c_seconds
        else:
            itr = 0

            while True:
                self.c_seconds, itr_state = h.helicsFederateRequestTimeIterative(
                    self.psse_federate, r_seconds, h.helics_iteration_request_iterate_if_needed
                )
                if itr_state == h.helics_iteration_result_next_step:
                    logger.debug("\tIteration complete!")
                    break

                error = max([abs(x["dStates"][0] - x["dStates"][1]) for k, x in self.subscriptions.items()])

                subscriptions = self.subscribe()
                for sub_name, sub_value in subscriptions.items():
                    if sub_name not in self.all_sub_results[self.sim.get_time()]:
                        self.all_sub_results[self.sim.get_time()][sub_name] = []
                    self.all_sub_results[self.sim.get_time()][sub_name].append(sub_value)

                self.sim.resolveStep(r_seconds)

                publications = self.publish()
                for pub_name, pub_value in publications.items():
                    if pub_name not in self.all_pub_results[self.sim.get_time()]:
                        self.all_pub_results[self.sim.get_time()][pub_name] = []
                    self.all_pub_results[self.sim.get_time()][pub_name].append(pub_value)

                itr += 1
                logger.debug(f"\titr = {itr}")

                if itr > self.settings.helics.max_coiterations or error < self.settings.helics.error_tolerance:
                    self.c_seconds, itr_state = h.helicsFederateRequestTimeIterative(
                        self.psse_federate, r_seconds, h.helics_iteration_request_no_iteration
                    )
                else:
                    pass

            return True, self.c_seconds

    def get_restructured_results(self, results: dict) -> dict:
        """result restructure for helics interface

        Args:
            results (dict): simulation results from result container

        Returns:
            dict: restructured result
        """

        results_dict = {}
        for k, d in results.items():
            c, p = k.split("_")
            if c not in results_dict:
                results_dict[c] = {}
            for n_raw, v in d.items():
                if isinstance(n_raw, str):
                    n = n_raw.replace(" ", "")
                else:
                    n = str(n_raw)
                if n not in results_dict[c]:
                    results_dict[c][n] = {}
                results_dict[c][n].update({p: v})
        return results_dict

    def publish(self) -> dict:
        """Publishes updated reslts each iteration

        Returns:
            dict: mapping of publication key to result
        """

        pub_results = {}
        for quantities, subsystem_buses in self.pub_struc:
            temp_res = self.sim.read_subsystems(quantities, subsystem_buses)
            temp_res = self.get_restructured_results(temp_res)
            for c_name, elm_info in temp_res.items():
                for name, v_info in elm_info.items():
                    for p_name, val in v_info.items():
                        pub_tag = f"{self.settings.helics.federate_name}.{c_name}.{name}.{p_name}"
                        pub_tag_reduced = f"{c_name}.{name}.{p_name}"
                        pub = self.publications[pub_tag]
                        dtype_matched = True
                        pub_results[pub_tag_reduced] = val
                        if isinstance(val, float):
                            h.helicsPublicationPublishDouble(pub, val)
                        elif isinstance(val, complex):
                            h.helicsPublicationPublishComplex(pub, val.real, val.imag)
                        elif isinstance(val, int):
                            h.helicsPublicationPublishInteger(pub, val)
                        elif isinstance(val, list):
                            h.helicsPublicationPublishVector(pub, val)
                        elif isinstance(val, str):
                            h.helicsPublicationPublishString(pub, val)
                        else:
                            dtype_matched = False
                            logger.warning(f"Publication {pub_tag} not updated")
                        if dtype_matched:
                            logger.debug(f"Publication {pub_tag} published: {val}")
        return pub_results

    def subscribe(self) -> dict:
        """Subscribes results each iteration and updates PSSE objects accordingly

        Returns:
            dict: mapping of subscription key to updated result
        """

        "Subscribes results each iteration and updates PSSE objects accordingly"
        for sub_tag, sub_data in self.subscriptions.items():
            if isinstance(sub_data["property"], str):
                sub_data["value"] = h.helicsInputGetDouble(sub_data["subscription"])
                self.psse_dict[sub_data["bus"]][sub_data["element_type"]][sub_data["element_id"]][
                    sub_data["property"]
                ] = (sub_data["value"], sub_data["scaler"])
            elif isinstance(sub_data["property"], list):
                sub_data["value"] = h.helicsInputGetVector(sub_data["subscription"])
                if isinstance(sub_data["value"], list) and len(sub_data["value"]) == len(sub_data["property"]):
                    for i, p in enumerate(sub_data["property"]):
                        self.psse_dict[sub_data["bus"]][sub_data["element_type"]][sub_data["element_id"]][p] = (
                            sub_data["value"][i],
                            sub_data["scaler"][i],
                        )

            logger.debug("Data received {} for tag {}".format(sub_data["value"], sub_tag))
            if self.settings.helics.iterative_mode:
                if self.c_seconds != self.c_seconds_old:
                    sub_data["dStates"] = [self.init_state] * self.n_states
                else:
                    sub_data["dStates"].insert(0, sub_data["dStates"].pop())
        all_values = {}
        for b, b_info in self.psse_dict.items():
            for t, t_info in b_info.items():
                for i, v_dict in t_info.items():
                    values = {}
                    j = 0
                    for p, v_raw in v_dict.items():
                        if isinstance(v_raw, tuple):
                            v, scale = v_raw
                            all_values[f"{t}.{b}.{i}.{p}"] = v
                            if isinstance(p, str):
                                ppty = f"realar{PROFILE_VALIDATION[t].index(p) + 1}"
                                values[ppty] = v * scale
                            elif isinstance(p, list):
                                for _, ppt in enumerate(p):
                                    ppty = f"realar{PROFILE_VALIDATION[t].index(ppt) + 1}"
                                    values[ppty] = v * scale
                        j += 1

                    is_empty = [0 if not vx else 1 for vx in values.values()]
                    if (
                        sum(is_empty) != 0
                        and sum(values.values()) < VALUE_UPDATE_BOUND
                        and sum(values.values()) > -VALUE_UPDATE_BOUND
                    ):
                        self.sim.update_object(t, b, i, values)
                        logger.debug(f"{t}.{b}.{i} = {values}")

                    else:
                        logger.debug("write failed")

        self.c_seconds_old = self.c_seconds
        return all_values

    def fill_missing_values(self, value):
        "fixes values before model dispatch"
        idx = [f"realar{PROFILE_VALIDATION[self.d_type].index(c) + 1}" for c in self.Columns]
        x = dict(zip(idx, list(value)))
        return x

    def __del__(self):
        "frees helics resources"
        h.helicsFederateFinalize(self.psse_federate)
        h.helicsFederateGetState(self.psse_federate)
        h.helicsFederateInfoFree(self.fedinfo)
        h.helicsFederateFree(self.psse_federate)
        logger.info("HELICS federate for PyDSS destroyed")



# --------------------------------------------------------------------------------------------------
# DC-AC Tool functions
# Bin Wang
# 4/28/2022
#
# Reference:
# Wang, Bin, and Jin Tan. 2022. DC-AC Tool: Fully Automating the Acquisition of AC Power Flow
# Solution. Golden, CO: National Renewable Energy Laboratory. NREL/TP-6A40-80100.
# https://www.nrel.gov/docs/fy22osti/80100.pdf
# --------------------------------------------------------------------------------------------------

import numpy as np


# --------------------------------------------------------------------------------------------------
# power flow data class
class PowerFlowData:
    def __init__(self, psse):
        self.psse = psse
        # bus data
        self.bus_num = []
        self.bus_type = []
        self.bus_Vpu = []

        # load data
        self.load_id = []
        self.load_bus = []
        self.load_Z = []
        self.load_I = []
        self.load_P = []
        self.load_MW = []
        self.load_Mvar = []
        self.load_MW_total = []
        self.load_Mvar_total = []

        # generator data
        self.gen_id = []
        self.gen_bus = []
        self.gen_status = []
        self.gen_S = []
        self.gen_mod = []
        self.gen_MW = []
        self.gen_Mvar = []
        self.gen_MW_total = []
        self.gen_Mvar_total = []

        # branch data
        self.brc_from = []
        self.brc_to = []
        self.brc_id = []
        self.brc_S = []
        self.brc_P = []
        self.brc_Q = []

    def getdata(self):
        # bus data
        self.bus_num = self.psse.abusint(-1, 2, "NUMBER")[1][0]
        self.bus_type = self.psse.abusint(-1, 2, "TYPE")[1][0]
        self.bus_Vpu = self.psse.abusreal(-1, 2, "PU")[1][0]

        # load data
        self.load_id = self.psse.aloadchar(-1, 4, "ID")[1][0]
        self.load_bus = self.psse.aloadint(-1, 4, "NUMBER")[1][0]
        self.load_Z = np.asarray(self.psse.aloadcplx(-1, 4, "YLACT")[1][0])
        self.load_I = np.asarray(self.psse.aloadcplx(-1, 4, "ILACT")[1][0])
        self.load_P = np.asarray(self.psse.aloadcplx(-1, 4, "MVAACT")[1][0])
        self.load_MW = self.load_Z.real + self.load_I.real + self.load_P.real
        self.load_Mvar = self.load_Z.imag + self.load_I.imag + self.load_P.imag
        self.load_MW_total = sum(self.load_MW)
        self.load_Mvar_total = sum(self.load_Mvar)

        # generator data
        self.gen_id = self.psse.amachchar(-1, 4, "ID")[1][0]
        self.gen_bus = self.psse.amachint(-1, 4, "NUMBER")[1][0]
        self.gen_status = self.psse.amachint(-1, 4, "STATUS")[1][0]
        self.gen_S = np.asarray(self.psse.amachcplx(-1, 4, "PQGEN")[1][0])
        self.gen_mod = np.asarray(self.psse.amachint(-1, 4, "WMOD")[1][0])
        self.gen_MW = self.gen_S.real
        self.gen_Mvar = self.gen_S.imag
        self.gen_MW_total = sum(self.gen_MW)
        self.gen_Mvar_total = sum(self.gen_Mvar)

        # branch data
        ierr, iarray = self.psse.abrnint(-1, 0, 0, 3, 2, ["FROMNUMBER", "TONUMBER"])
        self.brc_from = iarray[0][:]
        self.brc_to = iarray[1][:]
        self.brc_id = self.psse.abrnchar(-1, 0, 0, 3, 2, ["ID"])[1][0]
        self.brc_S = np.asarray(self.psse.abrncplx(-1, 1, 1, 3, 2, ["PQ"])[1][0])
        self.brc_P = self.brc_S.real
        self.brc_Q = self.brc_S.imag


# --------------------------------------------------------------------------------------------------
# check power flow solvability by running NR for up to multiple times (without adjusting loading)
# input: option - choose solver, n_flag1_max - maximum number of runs for power flow
# output: solved_flag, 0 - solved, 1 - reach maximum iteration number, 2 - blow up

# The following function was prepared based on this reference: Whit. 2012. 'Silencing PSSE Output,'
# Python for Power Systems. [Online] http://www.whit.com.au/blog/2012/03/silencing-psse-output/



"""Exceptions used in PyDSS"""


class InvalidConfigurationError(Exception):
    """Raised when a bad configuration is detected."""


class InvalidParameterError(Exception):
    """Raised when bad user input is detected."""


class HelicsConvergenceError(Exception):
    """Raised when OpenDSS fails to converge on a solution."""


class PsseConvergenceError(Exception):
    """Raised when OpenDSS fails to converge on a solution."""
